#!/usr/bin/python
#! --*-- utf-8 --*--

from vlcp.config import defaultconfig
from vlcp.server.module import Module,depend,call_api,api
from vlcp.event.runnable import RoutineContainer
from vlcp.utils.ethernet import ip4_addr
from vlcp.utils.dataobject import updater,\
            set_new,dump,ReferenceObject,\
            request_context, WeakReferenceObject
import vlcp.service.kvdb.objectdb as objectdb

from vlcp.utils.networkmodel import *
from vlcp.utils.netutils import ip_in_network, network_first, network_last,\
                        parse_ip4_address, parse_ip4_network,\
                        format_network_cidr

from uuid import uuid1
import copy
import logging
import itertools
from vlcp.utils.exceptions import AsyncTransactionLockException, WalkKeyNotRetrieved,\
    APIRejectedException
from collections import OrderedDict
from contextlib import suppress
from vlcp.utils.walkerlib import ensure_keys
from pychecktype.checked import checked
from vlcp.utils.typelib import ip_address_type, cidr_type, autoint, mac_address_type,\
    cidr_nonstrict_type
from pychecktype import tuple_, extra
from vlcp.utils.dhcp import dhcp_options_type

logger = logging.getLogger('viperflow')

#logger.setLevel(logging.DEBUG)

class UpdateConflictException(Exception):
    def __init__(self,desc="db update conflict"):
        super(UpdateConflictException,self).__init__(desc)


def dispatch_walker(parameter_dict, walker_map, create,
                    get_type):
    def _walker(walk, write, timestamp):
        # Collect type of each item and group into dict
        type_group = {}
        all_collected = True
        for key, parameters in parameter_dict.items():
            try:
                value = walk(key)
            except WalkKeyNotRetrieved:
                all_collected = False
            else:
                if value is None:
                    if not create:
                        raise ValueError(key + " does not exist")
                else:
                    if create:
                        # Raise exception
                        set_new(value, value)
                try:
                    type_, physicalnetwork_id = get_type(key, value, walk, parameters=parameters)
                except WalkKeyNotRetrieved:
                    all_collected = False
                else:
                    if type_ in type_group:
                        type_group[type_][key] = (parameters, physicalnetwork_id)
                    else:
                        type_group[type_] = {key: (parameters, physicalnetwork_id)}
        if not all_collected:
            return
        # Check if there are any types that are not in walker_map
        if any(t not in walker_map for t in type_group):
            raise AsyncTransactionLockException(type_group)
        for t, d in type_group.items():
            walker_map[t](walk, write, timestamp, {k: v[0] for k, v in d.items()})
    return _walker


def create_physicalnetwork_gettype(k, v, walk, parameters):
    return parameters['type'], parameters['id']


def physicalnetwork_gettype(k, v, walk, parameters):
    return v.type, v.id


def create_physicalport_gettype(k, v, walk, parameters):
    phynet = walk(PhysicalNetwork.default_key(parameters['physicalnetwork']))
    if phynet is None:
        raise ValueError("Physical network " + parameters['physicalnetwork'] + ' not exists')
    return phynet.type, phynet.id


def create_physicalport_prekey(key, parameters):
    return (PhysicalNetwork.default_key(parameters['physicalnetwork']),)


def physicalport_gettype(k, v, walk, parameters):
    phynet = walk(v.physicalnetwork.getkey())
    return phynet.type, phynet.id
    

create_logicalnetwork_gettype = create_physicalport_gettype


create_logicalnetwork_prekey = create_physicalport_prekey


logicalnetwork_gettype = physicalport_gettype

lease_time_type = (autoint, extra(str, check=lambda x: x == 'infinite'), None)

def dispatch_async_walker(parameter_dict, create, get_type, publicapi_,
                          direct_get_type = False,
                          pre_keys = None):
    async def _asyncwalker(last_info, container):
        walker_map = {}
        keys = set(parameter_dict)
        if last_info is None and direct_get_type:
            type_group = {}
            for key, parameters in parameter_dict.items():
                type_, physicalnetwork_id = get_type(key, None, None, parameters=parameters)
                if type_ in type_group:
                    type_group[type_][key] = (parameters, physicalnetwork_id)
                else:
                    type_group[type_] = {key: (parameters, physicalnetwork_id)}
            last_info = type_group
        if last_info is not None:            
            for t, d in last_info.items():
                # Collect walker and keys from public API
                try:
                    walker, k_ = await call_api(container, 'public', publicapi_,
                                                {'type': t})
                except APIRejectedException:
                    raise ValueError("Physical network type %r is not supported, or the corresponding network plugin is not loaded" % (t,))
                for _, (parameters, phynet_id) in d.items():
                    # If it is not changed, this network is needed
                    # this reduces an extra retry
                    keys.add(PhysicalNetwork.default_key(phynet_id))
                if k_:
                    for _, (parameters, phynet_id) in d.items():
                        keys.update(k_(phynet_id, parameters))
                walker_map[t] = walker
        else:
            if pre_keys:
                for key, parameters in parameter_dict.items():
                    keys.update(pre_keys(key, parameters))
        return (tuple(keys), dispatch_walker(parameter_dict, walker_map, create, get_type))
    return _asyncwalker


@defaultconfig
@depend(objectdb.ObjectDB)
class ViperFlow(Module):
    """
    Standard network model for L2 SDN
    """
    def __init__(self,server):
        super(ViperFlow,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        self._reqid = 0
        self.createAPI(api(self.createphysicalnetwork,self.app_routine),
                       api(self.createphysicalnetworks,self.app_routine),
                       api(self.updatephysicalnetwork,self.app_routine),
                       api(self.updatephysicalnetworks,self.app_routine),
                       api(self.deletephysicalnetwork,self.app_routine),
                       api(self.deletephysicalnetworks,self.app_routine),
                       api(self.listphysicalnetworks,self.app_routine),
                       api(self.createphysicalport,self.app_routine),
                       api(self.createphysicalports,self.app_routine),
                       api(self.updatephysicalport,self.app_routine),
                       api(self.updatephysicalports,self.app_routine),
                       api(self.deletephysicalport,self.app_routine),
                       api(self.deletephysicalports,self.app_routine),
                       api(self.listphysicalports,self.app_routine),
                       api(self.createlogicalnetwork,self.app_routine),
                       api(self.createlogicalnetworks,self.app_routine),
                       api(self.updatelogicalnetwork,self.app_routine),
                       api(self.updatelogicalnetworks,self.app_routine),
                       api(self.deletelogicalnetwork,self.app_routine),
                       api(self.deletelogicalnetworks,self.app_routine),
                       api(self.listlogicalnetworks,self.app_routine),
                       api(self.createlogicalport,self.app_routine),
                       api(self.createlogicalports,self.app_routine),
                       api(self.updatelogicalport,self.app_routine),
                       api(self.updatelogicalports,self.app_routine),
                       api(self.deletelogicalport,self.app_routine),
                       api(self.deletelogicalports,self.app_routine),
                       api(self.listlogicalports,self.app_routine),
                       api(self.createsubnet,self.app_routine),
                       api(self.createsubnets,self.app_routine),
                       api(self.updatesubnet,self.app_routine),
                       api(self.updatesubnets,self.app_routine),
                       api(self.deletesubnet,self.app_routine),
                       api(self.deletesubnets,self.app_routine),
                       api(self.listsubnets,self.app_routine)
                       )

    async def _dumpkeys(self, keys, filter=None):
        self._reqid += 1
        reqid = ('viperflow', self._reqid)

        with request_context(reqid, self.app_routine):
            retobjs = await call_api(self.app_routine,'objectdb','mget',{'keys':keys,'requestid':reqid})
            if filter is None:
                return [dump(o) for o in retobjs]
            else:
                return [dump(o) for o in retobjs if o is not None and all(getattr(o, k, None) == v for k, v in filter.items())]

    async def _dumpone(self, key, filter):
        return await self._dumpkeys([key], filter)
    
    async def createphysicalnetwork(self, type: str = 'vlan', id: (str, None) = None, **kwargs: {'?vnirange': [tuple_((int, int))],
                                                                                                 "?vlanrange": [tuple_((int, int))]}):
        """
        Create physical network.
        
        :param type: Network type, usually one of *vlan*, *vxlan*, *local*, *native*
        
        :param id: Specify the created physical network ID. If omitted or None, an UUID is
                   generated.
        
        :param \*\*kwargs: extended creation parameters. Look for the document of the corresponding
                           driver. Common options include:
                           
                           vnirange
                              list of ``[start,end]`` ranges like ``[[1000,2000]]``. Both *start* and
                              *end* are included. It specifies the usable VNI ranges for VXLAN network.
                              
                           vlanrange
                              list of ``[start,end]`` ranges like ``[[1000,2000]]``. Both *start* and
                              *end* are included. It specifies the usable VLAN tag ranges for VLAN network.
        
        :return: A dictionary of information of the created physical network.
        """
        if not id:
            id = str(uuid1())

        network = {'type':type,'id':id}
        network.update(kwargs)

        return await self.createphysicalnetworks([network])
    
    @checked
    async def createphysicalnetworks(self,networks: [{"?id": str, "type": str,
                                                      '?vnirange': [tuple_((int, int))],
                                                      "?vlanrange": [tuple_((int, int))]}]):
        """
        Create multiple physical networks in a transaction.
        
        :param networks: each should be a dictionary contains all the parameters in ``createphysicalnetwork``
        
        :return: A list of dictionaries of information of the created physical networks.
        """
        #networks [{type='vlan' or 'vxlan',id = None or uuid1(),'vlanrange':[(100,200),(400,401)],kwargs}]
        parameter_dict = OrderedDict()
        # first check id is None, allocate for it
        # group by type, do it use type driver
        for network in networks:
            #
            # deepcopy every networks elements
            # case:[network]*N point to same object will auto create same id
            #
            network = copy.deepcopy(network)
            if 'id' not in network:
                network['id'] = str(uuid1())
            phynetkey = PhysicalNetwork.default_key(network['id'])
            if phynetkey in parameter_dict:
                raise ValueError("Repeated ID: " + network['id'])
            parameter_dict[phynetkey] = network
        await call_api(self.app_routine, 'objectdb', 'asyncwritewalk',
                        {"asyncwalker": dispatch_async_walker(parameter_dict, True, create_physicalnetwork_gettype,
                                                              'createphysicalnetwork',
                                                              direct_get_type=True),
                         "withtime": True})
        return await self._dumpkeys(parameter_dict)
    
    async def updatephysicalnetwork(self, id: str, **kwargs: {'?vnirange': [tuple_((int, int))],
                                                              "?vlanrange": [tuple_((int, int))]}):
        """
        Update physical network with the specified ID.
        
        :param id: physical network ID
        
        :param \*\*kwargs: attributes to be updated, usually the same attributes for creating.
        
        :return: A dictionary of information of the updated physical network.        
        """
        network = {"id":id}
        network.update(kwargs)

        return await self.updatephysicalnetworks([network])
    
    @checked
    async def updatephysicalnetworks(self,networks: [{"id": str,
                                                      '?vnirange': [tuple_((int, int))],
                                                      "?vlanrange": [tuple_((int, int))]}]):
        """
        Update multiple physical networks in a transaction
        
        :param networks: a list of dictionaries, each contains parameters of ``updatephysicalnetwork``
        
        :return: A list of dictionaries of information of the updated physical network.
        """

        # networks [{"id":phynetid,....}]

        parameter_dict = OrderedDict()
        for network in networks:
            if 'type' in network:
                raise ValueError("physical network type can't be changed")

            phynetkey = PhysicalNetwork.default_key(network['id'])
            if phynetkey in parameter_dict:
                raise ValueError("Repeated ID: "+network['id'])
            parameter_dict[phynetkey] = network

        await call_api(self.app_routine, 'objectdb', 'asyncwritewalk',
                        {"asyncwalker": dispatch_async_walker(parameter_dict, False, physicalnetwork_gettype,
                                                              'updatephysicalnetwork'),
                         "withtime": True})
        return await self._dumpkeys(parameter_dict)
    
    async def deletephysicalnetwork(self, id: str):
        """
        Delete physical network with specified ID
        
        :param id: Physical network ID
        
        :return: ``{"status": "OK"}``
        """
        network = {"id":id}
        return await self.deletephysicalnetworks([network])
    
    @checked
    async def deletephysicalnetworks(self,networks: [{"id": str}]):
        """
        Delete multiple physical networks with a transaction
        
        :param networks: a list of ``{"id": <id>}`` dictionaries.
        
        :return: ``{"status": "OK"}``
        """
        # networks [{"id":id},{"id":id}]
        parameter_dict = {}
        for network in networks:
            phynetkey = PhysicalNetwork.default_key(network['id'])
            if phynetkey in parameter_dict:
                raise ValueError("Repeated ID: "+network['id'])
            parameter_dict[phynetkey] = network

        await call_api(self.app_routine, 'objectdb', 'asyncwritewalk',
                {"asyncwalker": dispatch_async_walker(parameter_dict, False, physicalnetwork_gettype,
                                                      'deletephysicalnetwork'),
                 "withtime": True})
        return {"status":'OK'}

    async def listphysicalnetworks(self,id = None,**kwargs):
        """
        Query physical network information
        
        :param id: If specified, only return the physical network with the specified ID.
        
        :param \*\*kwargs: customized filters, only return a physical network if
                           the attribute value of this physical network matches
                           the specified value.
        
        :return: A list of dictionaries each stands for a matched physical network
        """
        def set_walker(key,set,walk,save):
            if set is None:
                return
            for refnetwork in set.dataset():
                networkkey = refnetwork.getkey()
                with suppress(WalkKeyNotRetrieved):
                    networkobj = walk(networkkey)
                    if all(getattr(networkobj,k,None) == v for k,v in kwargs.items()):
                        save(networkkey)

        def walker_func(set_func):
            def walker(key,obj,walk,save):
                if obj is None:
                    return
                set_walker(key,set_func(obj),walk,save)
            return walker
        # get all phynet
        if not id:
            physetkey = PhysicalNetworkSet.default_key()
            # an unique id used to unwatch
            self._reqid += 1
            reqid = ('viperflow',self._reqid)
            with request_context(reqid, self.app_routine):
                _, values = await call_api(self.app_routine,'objectdb','walk',{'keys':[physetkey],
                                                'walkerdict':{physetkey:walker_func(lambda x:x.set)},
                                                'requestid':reqid})
                return [dump(r) for r in values]
        else:
            # get that id phynet info
            phynetkey = PhysicalNetwork.default_key(id)
            return await self._dumpone(phynetkey,kwargs)
    
    async def createphysicalport(self,physicalnetwork: str, name: str, vhost: str='',
                                 systemid: str='%',
                                 bridge: str='%',
                                 **kwargs):
        """
        Create physical port
        
        :param physicalnetwork: physical network this port is in.
        
        :param name: port name of the physical port, should match the name in OVSDB
        
        :param vhost: only match ports for the specified vHost
        
        :param systemid: only match ports on this systemid; or '%' to match all systemids.
        
        :param bridge: only match ports on bridges with this name; or '%' to match all bridges.
        
        :param \*\*kwargs: customized creation options, check the driver document
        
        :return: A dictionary containing information of the created physical port.
        """
        port = {'physicalnetwork':physicalnetwork,'name':name,'vhost':vhost,'systemid':systemid,'bridge':bridge}
        port.update(kwargs)

        return await self.createphysicalports([port])
    
    @checked
    async def createphysicalports(self,ports: [{"physicalnetwork": str,
                                                "name": str,
                                                "?vhost": str,
                                                "?systemid": str,
                                                "?bridge": str}]):
        """
        Create multiple physical ports in a transaction
        
        :param ports: A list of dictionaries, each contains all parameters for ``createphysicalport``
        
        :return: A list of dictionaries of information of the created physical ports
        """
        # ports [{'physicalnetwork':id,'name':eth0,'vhost':'',systemid:'%'},{.....}]

        parameter_dict = OrderedDict()

        for port in ports:
            port = copy.deepcopy(port)
            port.setdefault('vhost','')
            port.setdefault('systemid','%')
            port.setdefault('bridge','%')
            
            key = PhysicalPort.default_key(port['vhost'], port['systemid'], port['bridge'], port['name'])
            if key in parameter_dict:
                raise ValueError("Repeated key: " + '.'.join([port['vhost'],port['systemid'],port['bridge'],port['name']]))
            parameter_dict[key] = port
        
        await call_api(self.app_routine, 'objectdb', 'asyncwritewalk',
                        {"asyncwalker": dispatch_async_walker(parameter_dict, True, create_physicalport_gettype,
                                                              'createphysicalport',
                                                              pre_keys=create_physicalport_prekey),
                         "withtime": True})
        return await self._dumpkeys(parameter_dict)
    
    async def updatephysicalport(self, name: str, vhost : str = '',
                                       systemid: str = '%',
                                       bridge: str = '%',
                                       **args):
        """
        Update physical port
        
        :param name: Update physical port with this name.
        
        :param vhost: Update physical port with this vHost.
        
        :param systemid: Update physical port with this systemid.
        
        :param bridge: Update physical port with this bridge name.
        
        :param \*\*kwargs: Attributes to be updated
        
        :return: Updated result as a dictionary.
        """
        port = {'name':name,'vhost':vhost,'systemid':systemid,'bridge':bridge}
        port.update(args)

        return await self.updatephysicalports([port])
    
    @checked
    async def updatephysicalports(self, ports: [{"name": str,
                                                "?vhost": str,
                                                "?systemid": str,
                                                "?bridge": str}]):
        """
        Update multiple physical ports with a transaction
        
        :param ports: a list of ``updatephysicalport`` parameters
                
        :return: Updated result as a list of dictionaries.
        """
        # ports [{'name':eth0,'vhost':'',systemid:'%'},{.....}]
        parameter_dict = OrderedDict()
        for port in ports:
            if 'physicalnetwork' in port:
                raise ValueError("physical network cannot be changed")
            port = copy.deepcopy(port)
            port.setdefault("vhost","")
            port.setdefault("systemid","%")
            port.setdefault("bridge","%")
            portkey = PhysicalPort.default_key(port['vhost'],
                        port['systemid'], port['bridge'], port['name'])
            
            if portkey in parameter_dict:
                raise ValueError("Repeated key: " + '.'.join([port['vhost'],port['systemid'],port['bridge'],port['name']]))
            parameter_dict[portkey] = port
        await call_api(self.app_routine, 'objectdb', 'asyncwritewalk',
                        {"asyncwalker": dispatch_async_walker(parameter_dict, False, physicalport_gettype,
                                                              'updatephysicalport'),
                         "withtime": True})
        return await self._dumpkeys(parameter_dict)
    
    async def deletephysicalport(self, name: str, vhost : str = '',
                                       systemid: str = '%',
                                       bridge: str = '%'):
        """
        Delete a physical port
        
        :param name: physical port name.
        
        :param vhost: physical port vHost.
        
        :param systemid: physical port systemid.
        
        :param bridge: physcial port bridge.
        
        :return: ``{"status": "OK"}``
        """
        port = {'name':name,'vhost':vhost,'systemid':systemid,'bridge':bridge}
        return await self.deletephysicalports([port])

    @checked
    async def deletephysicalports(self, ports: [{"name": str,
                                                  "?vhost": str,
                                                  "?systemid": str,
                                                  "?bridge": str}]):
        """
        Delete multiple physical ports in a transaction
        
                Delete a physical port
        
        :param ports: a list of ``deletephysicalport`` parameters
        
        :return: ``{"status": "OK"}``
        """

        parameter_dict = {}
        for port in ports:
            port = copy.deepcopy(port)
            port.setdefault("vhost","")
            port.setdefault("systemid","%")
            port.setdefault("bridge","%")
            portkey = PhysicalPort.default_key(port['vhost'],
                        port['systemid'], port['bridge'], port['name'])
            
            if portkey in parameter_dict:
                raise ValueError("Repeated key: " + '.'.join([port['vhost'],port['systemid'],port['bridge'],port['name']]))
            parameter_dict[portkey] = port
        await call_api(self.app_routine, 'objectdb', 'asyncwritewalk',
                        {"asyncwalker": dispatch_async_walker(parameter_dict, False, physicalport_gettype,
                                                              'deletephysicalport'),
                         "withtime": True})
        return {"status":'OK'}

    async def listphysicalports(self,name = None,physicalnetwork = None,vhost='',
            systemid='%',bridge='%',**kwargs):
        """
        Query physical port information
        
        :param name: If specified, only return the physical port with the specified name.
        
        :param physicalnetwork: If specified, only return physical ports in that physical network
        
        :param vhost: If specified, only return physical ports for that vHost.
        
        :param systemid: If specified, only return physical ports for that systemid.
        
        :param bridge: If specified, only return physical ports for that bridge. 
        
        :param \*\*kwargs: customized filters, only return a physical network if
                           the attribute value of this physical network matches
                           the specified value.
        
        :return: A list of dictionaries each stands for a matched physical network
        """

        if name:
            phyportkey = PhysicalPort.default_key(vhost,systemid,bridge,name)
            return await self._dumpone(phyportkey,kwargs)
        else:
            if physicalnetwork:
                # specify physicalnetwork , find it in map , filter it ..
                physical_map_key = PhysicalNetworkMap.default_key(physicalnetwork)

                self._reqid += 1
                reqid = ('viperflow',self._reqid)
                with request_context(reqid, self.app_routine):
                    def walk_map(key,value,walk,save):
                        if value is None:
                            return
    
                        for weakobj in value.ports.dataset():
                            phyport_key = weakobj.getkey()
    
                            with suppress(WalkKeyNotRetrieved):
                                phyport_obj = walk(phyport_key)
                                if all(getattr(phyport_obj,k,None) == v for k ,v in kwargs.items()):
                                    save(phyport_key)
                    _, values = await call_api(self.app_routine, 'objectdb', 'walk',
                                               {"keys":[physical_map_key],
                                                "walkerdict":{physical_map_key:walk_map},
                                                "requestid":reqid})

                    return [dump(r) for r in values]

            else:
                # find it in all, , filter it ,,
                phyport_set_key = PhysicalPortSet.default_key()

                self._reqid += 1
                reqid = ('viperflow',self._reqid)
                with request_context(reqid, self.app_routine):
                    def walk_set(key,value,walk,save):
                        if value is None:
                            return
    
                        for weakobj in value.set.dataset():
                            phyport_key = weakobj.getkey()
    
                            with suppress(WalkKeyNotRetrieved):
                                phyport_obj = walk(phyport_key)
                                if all(getattr(phyport_obj,k,None) == v for k,v in kwargs.items()):
                                    save(phyport_key)
    
                    _, values = await call_api(self.app_routine,'objectdb','walk',{"keys":[phyport_set_key],
                                                                "walkerdict":{phyport_set_key:walk_set},
                                                                "requestid":reqid})
    
                    return [dump(r) for r in values]
    
    async def createlogicalnetwork(self, physicalnetwork: str,
                                         id: (str, None) = None,
                                         **kwargs: {"?vni": autoint,
                                                    "?vxlan": autoint,
                                                    "?mtu": autoint,
                                                    "?dns_nameservers": ([ip_address_type], None),
                                                    "?domain_name": (str, None),
                                                    "?ntp_servers": ([ip_address_type], None),
                                                    "?lease_time": lease_time_type,
                                                    "?extra_dhcp_options": dhcp_options_type
                                                    }):
        """
        Create logical network
        
        :param physicalnetwork: physical network ID that contains this logical network
        
        :param id: logical network ID. If ommited an UUID is generated.
        
        :param \*\*kwargs: customized options for logical network creation.
                           Common options include:
                           
                           vni/vxlan
                              Specify VNI / VLAN tag for VXLAN / VLAN network. If omitted,
                              an unused VNI / VLAN tag is picked automatically.
                           
                           mtu
                              MTU value for this network. You can use 1450 for VXLAN networks.
        
        :return: A dictionary of information of the created logical port
        """
        if not id:
            id = str(uuid1())
        network = {'physicalnetwork':physicalnetwork,'id':id}
        network.update(kwargs)

        return await self.createlogicalnetworks([network])
    
    @checked
    async def createlogicalnetworks(self,networks: [{"physicalnetwork": str,
                                                     "?id": str,
                                                     "?vni": autoint,
                                                     "?vxlan": autoint,
                                                     "?mtu": autoint,
                                                     "?dns_nameservers": ([ip_address_type], None),
                                                     "?domain_name": (str, None),
                                                     "?ntp_servers": ([ip_address_type], None),
                                                     "?lease_time": lease_time_type,
                                                     "?extra_dhcp_options": dhcp_options_type}]):
        """
        Create multiple logical networks in a transaction.
        
        :param networks: a list of ``createlogicalnetwork`` parameters.
        
        :return: a list of dictionaries for the created logical networks.
        """
        # networks [{'physicalnetwork':'id','id':'id' ...},{'physicalnetwork':'id',...}]

        parameter_dict = OrderedDict()

        for network in networks:
            network = copy.deepcopy(network)
            if 'id' not in network:
                network['id'] = str(uuid1())
            key = LogicalNetwork.default_key(network['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: " + network['id'])
            parameter_dict[key] = network        
        await call_api(self.app_routine, 'objectdb', 'asyncwritewalk',
                        {"asyncwalker": dispatch_async_walker(parameter_dict, True, create_logicalnetwork_gettype,
                                                              'createlogicalnetwork',
                                                              pre_keys=create_logicalnetwork_prekey),
                         "withtime": True})
        return await self._dumpkeys(parameter_dict)
    
    async def updatelogicalnetwork(self, id: str, **kwargs: {"?vni": autoint,
                                                            "?vxlan": autoint,
                                                            "?mtu": autoint,
                                                            "?dns_nameservers": ([ip_address_type], None),
                                                            "?domain_name": (str, None),
                                                            "?ntp_servers": ([ip_address_type], None),
                                                            "?lease_time": lease_time_type,
                                                            "?extra_dhcp_options": dhcp_options_type}):
        """
        Update logical network attributes of the ID
        """
        # update phynetid is disabled
        network = {'id':id}
        network.update(kwargs)

        return await self.updatelogicalnetworks([network])
    
    @checked
    async def updatelogicalnetworks(self,networks: [{"id": str,
                                                     "?vni": autoint,
                                                     "?vxlan": autoint,
                                                     "?mtu": autoint,
                                                     "?dns_nameservers": ([ip_address_type], None),
                                                     "?domain_name": (str, None),
                                                     "?ntp_servers": ([ip_address_type], None),
                                                     "?lease_time": lease_time_type,
                                                     "?extra_dhcp_options": dhcp_options_type}]):
        """
        Update multiple logical networks in a transaction
        """
        #networks [{'id':id,....},{'id':id,....}]

        parameter_dict = OrderedDict()

        for network in networks:
            if 'physicalnetwork' in network:
                raise ValueError("physical network cannot be changed")
            key = LogicalNetwork.default_key(network['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: " + network['id'])
            parameter_dict[key] = network
        await call_api(self.app_routine, 'objectdb', 'asyncwritewalk',
                        {"asyncwalker": dispatch_async_walker(parameter_dict, False, logicalnetwork_gettype,
                                                              'updatelogicalnetwork'),
                         "withtime": True})
        return await self._dumpkeys(parameter_dict)
    
    async def deletelogicalnetwork(self, id: str):
        """
        Delete logical network
        """
        network = {'id':id}
        return await self.deletelogicalnetworks([network])
    
    @checked
    async def deletelogicalnetworks(self, networks: [{"id": str}]):
        """
        Delete logical networks
        
        :param networks: a list of ``{"id":id}``
        
        :return: ``{"status": "OK"}``
        """
        # networks [{"id":id},{"id":id}]
        parameter_dict = OrderedDict()

        for network in networks:
            key = LogicalNetwork.default_key(network['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: " + network['id'])
            parameter_dict[key] = network
        await call_api(self.app_routine, 'objectdb', 'asyncwritewalk',
                        {"asyncwalker": dispatch_async_walker(parameter_dict, False, logicalnetwork_gettype,
                                                              'deletelogicalnetwork'),
                         "withtime": True})
        return {"status":'OK'}

    async def listlogicalnetworks(self,id = None,physicalnetwork = None,**kwargs):
        """
        Query logical network information
        
        :param id: If specified, only return the logical network with the specified ID.
        
        :param physicalnetwork: If specified, only return logical networks in this physical network.
        
        :param \*\*kwargs: customized filters, only return a logical network if
                           the attribute value of this logical network matches
                           the specified value.
        
        :return: A list of dictionaries each stands for a matched logical network
        """

        if id:
            # specify id ,, find it,, filter it
            lgnetkey = LogicalNetwork.default_key(id)

            return await self._dumpone(lgnetkey,kwargs)

        else:
            if physicalnetwork:
                # specify physicalnetwork , find it in physicalnetwork,  filter it
                self._reqid += 1
                reqid = ('viperflow',self._reqid)

                physicalnetwork_map_key = PhysicalNetworkMap.default_key(physicalnetwork)

                def walk_map(key,value,walk,save):
                    if value is None:
                        return

                    for weakobj in value.logicnetworks.dataset():
                        lgnet_key = weakobj.getkey()

                        with suppress(WalkKeyNotRetrieved):
                            lgnet_obj = walk(lgnet_key)
                            if all(getattr(lgnet_obj,k,None) == v for k ,v in kwargs.items()):
                                save(lgnet_key)
                with request_context(reqid, self.app_routine):
                    _, values = await call_api(self.app_routine,'objectdb','walk',
                                                    {'keys':[physicalnetwork_map_key],
                                                    'walkerdict':{physicalnetwork_map_key:walk_map},
                                                    'requestid':reqid})
                    return [dump(r) for r in values]
            else:
                # find it in all set , filter it
                self._reqid += 1
                reqid = ('viperflow',self._reqid)
                lgnet_set_key = LogicalNetworkSet.default_key()

                def walk_set(key,value,walk,save):
                    if value is None:
                        return

                    for weakobj in value.set.dataset():
                        lgnet_key = weakobj.getkey()

                        with suppress(WalkKeyNotRetrieved):
                            lgnet_obj = walk(lgnet_key)
                            if all(getattr(lgnet_obj,k,None) == v for k,v in kwargs.items()):
                                save(lgnet_key)
                with request_context(reqid, self.app_routine):
                    _, values = await call_api(self.app_routine,"objectdb","walk",
                                                    {'keys':[lgnet_set_key],
                                                     "walkerdict":{lgnet_set_key:walk_set},"requestid":reqid})
                    return [dump(r) for r in values]
    
    async def createlogicalport(self, logicalnetwork: str,
                                      id: (str, None) = None,
                                      subnet: (str, None) = None,
                                      **kwargs: {"?mac_address": mac_address_type,
                                                 "?ip_address": ip_address_type,
                                                 "?hostname": (str, None),
                                                 "?extra_dhcp_options": dhcp_options_type}):
        """
        Create logical port
        
        :param logicalnetwork: logical network containing this port
        
        :param id: logical port id. If omitted an UUID is created.
        
        :param subnet: subnet containing this port
        
        :param \*\*kwargs: customized options for creating logical ports.
                           Common options are:
                           
                           mac_address
                              port MAC address
                           
                           ip_address
                              port IP address
        
        :return: a dictionary for the logical port
        """
        if not id:
            id = str(uuid1())
        
        if subnet:
            port = {'logicalnetwork':logicalnetwork,'id':id,'subnet':subnet}
        else:
            port = {'logicalnetwork':logicalnetwork,'id':id}
        
        port.update(kwargs)

        return await self.createlogicalports([port])
    
    @checked
    async def createlogicalports(self, ports: [{"?id": str,
                                                "logicalnetwork": str,
                                                "?subnet": str,
                                                "?mac_address": mac_address_type,
                                                "?ip_address": ip_address_type,
                                                "?hostname": (str, None),
                                                "?extra_dhcp_options": dhcp_options_type}]):
        """
        Create multiple logical ports in a transaction
        """
        parameter_dict = OrderedDict()
        keys = set()
        for port in ports:
            port = copy.deepcopy(port)
            if 'id' not in port:
                port['id'] = str(uuid1())
            key = LogicalPort.default_key(port['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: "+ port['id'])
            if 'logicalnetwork' not in port:
                raise ValueError("must specify logicalnetwork ID")
            keys.add(key)
            keys.add(LogicalNetwork.default_key(port['logicalnetwork']))
            keys.add(LogicalNetworkMap.default_key(port['logicalnetwork']))
            if 'subnet' in port:
                keys.add(SubNet.default_key(port['subnet']))
                keys.add(SubNetMap.default_key(port['subnet']))
            parameter_dict[key] = port
        keys.add(LogicalPortSet.default_key())
        def walker(walk, write):
            # Process logical ports with specified IP address first,
            # so the automatically allocated IPs do not conflict
            # with specified IPs
            for key, parameters in sorted(parameter_dict.items(),
                                          key=lambda x: 'ip_address' in x[1],
                                          reverse=True):
                with suppress(WalkKeyNotRetrieved):
                    value = walk(key)
                    value = set_new(value, LogicalPort.create_instance(parameters['id']))
                    with suppress(WalkKeyNotRetrieved):
                        lognet_id = parameters['logicalnetwork']
                        lognet = walk(LogicalNetwork.default_key(lognet_id))
                        if not lognet:
                            raise ValueError("Logical network " + lognet_id + " not exists")
                        value.network = lognet.create_reference()
                        logmap = walk(LogicalNetworkMap.default_key(lognet_id))
                        logmap.ports.dataset().add(value.create_weakreference())
                        write(logmap.getkey(), logmap)
                        if 'subnet' in parameters:
                            subnet_id = parameters['subnet']
                            subnet = walk(SubNet.default_key(subnet_id))
                            if not subnet:
                                raise ValueError("Subnet " + subnet_id + " not exists")
                            if subnet.create_weakreference() not in logmap.subnets.dataset():
                                raise ValueError("Specified subnet " + subnet_id + " is not in logical network " + lognet_id)
                            subnet_map = walk(SubNetMap.default_key(subnet_id))
                            value.subnet = subnet.create_reference()
                            if 'ip_address' in parameters:
                                ip_address = parse_ip4_address(parameters['ip_address'])
                                value.ip_address = ip4_addr.formatter(ip_address)
                                # check ip_address in cidr
                                start = parse_ip4_address(subnet.allocated_start)
                                end = parse_ip4_address(subnet.allocated_end)
                                try:
                                    assert start <= ip_address <= end
                                    if hasattr(subnet, 'gateway'):
                                        assert ip_address != parse_ip4_address(subnet.gateway)
                                except Exception:
                                    raise ValueError("Specified ip_address " + parameters['ip_address'] + " is not an usable IP address in subnet " + subnet_id)

                                if str(ip_address) not in subnet_map.allocated_ips:
                                    subnet_map.allocated_ips[str(ip_address)] = value.create_weakreference()
                                else:
                                    raise ValueError("IP address " + parameters['ip_address'] + " has been used in subnet " + subnet_id)
                            else:
                                # allocated ip_address from cidr
                                start = parse_ip4_address(subnet.allocated_start)
                                end = parse_ip4_address(subnet.allocated_end)
                                gateway = None
                                if hasattr(subnet, "gateway"):
                                    gateway = parse_ip4_address(subnet.gateway)
    
                                for ip_address in range(start, end + 1):
                                    if str(ip_address) not in subnet_map.allocated_ips and ip_address != gateway:
                                        value.ip_address = ip4_addr.formatter(ip_address)
                                        subnet_map.allocated_ips[str(ip_address)] = value.create_weakreference()
                                        break
                                else:
                                    raise ValueError("Cannot allocate an available IP address from subnet " + subnet_id)
                            write(subnet_map.getkey(), subnet_map)
                        # Process other parameters
                        for k,v in parameters.items():
                            if k not in ('id', 'logicalnetwork', 'subnet', 'ip_address'):
                                setattr(value, k, v)
                        write(key, value)
                    with suppress(WalkKeyNotRetrieved):
                        logport_set = walk(LogicalPortSet.default_key())
                        logport_set.set.dataset().add(value.create_weakreference())
                        write(logport_set.getkey(), logport_set)
        await call_api(self.app_routine, 'objectdb', 'writewalk', {'keys': keys, 'walker': walker})
        return await self._dumpkeys(parameter_dict)

    async def updatelogicalport(self, id: str, **kwargs: {"?mac_address": mac_address_type,
                                                          "?ip_address": ip_address_type,
                                                          "?hostname": (str, None),
                                                          "?extra_dhcp_options": dhcp_options_type}):
        "Update attributes of the specified logical port"
        if not id :
            raise ValueError("must specify id")

        port = {"id":id}
        port.update(kwargs)

        return await self.updatelogicalports([port])
    
    @checked
    async def updatelogicalports(self, ports: [{"id": str,
                                                "?mac_address": mac_address_type,
                                                "?ip_address": ip_address_type,
                                                "?hostname": (str, None),
                                                "?extra_dhcp_options": dhcp_options_type}]):
        "Update multiple logcial ports"
        # ports [{"id":id,...},{...}]
        
        parameter_dict = OrderedDict()
        for port in ports:
            port = copy.deepcopy(port)
            key = LogicalPort.default_key(port['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: " + port['id'])
            if 'logicalnetwork' in port:
                raise ValueError("logical network cannot be changed")
            if 'network' in port:
                raise ValueError("logical network cannot be changed")
            if 'subnet' in port:
                raise ValueError("subnet cannot be changed")
            parameter_dict[key] = port

        def walker(walk, write):
            # Must deallocate all IP addresses before allocating new
            deallocated_all = True
            for key, parameters in parameter_dict.items():
                try:
                    value = walk(key)
                    if value is None:
                        raise ValueError("Logical port " + parameters['id'] + " not exists")
                    if 'ip_address' in parameters and hasattr(value, 'subnet') and hasattr(value, 'ip_address'):
                        # Subnet is needed when allocating IP address
                        ensure_keys(walk, value.subnet.getkey())
                        subnet_map = walk(SubNetMap._subnet.leftkey(value.subnet))
                        del subnet_map.allocated_ips[str(parse_ip4_address(value.ip_address))]
                        write(subnet_map.getkey(), subnet_map)
                except WalkKeyNotRetrieved:
                    deallocated_all = False
            if not deallocated_all:
                return
            # Update processing
            for key, parameters in parameter_dict.items():
                value = walk(key)
                if 'ip_address' in parameters and hasattr(value, 'subnet'):
                    with suppress(WalkKeyNotRetrieved):
                        ensure_keys(walk, value.subnet.getkey(),
                                          SubNetMap._subnet.leftkey(value.subnet))
                        try:
                            subnet = walk(value.subnet.getkey())
                        except WalkKeyNotRetrieved:
                            # Also retrieve subnet map to prevent another try
                            ensure_keys(walk, SubNetMap._subnet.leftkey(value.subnet))
                            raise
                        subnet_map = walk(SubNetMap._subnet.leftkey(value.subnet))
                        ip_address = parse_ip4_address(parameters['ip_address'])
                        start = parse_ip4_address(subnet.allocated_start)
                        end = parse_ip4_address(subnet.allocated_end)
                        try:
                            assert start <= ip_address <= end
                            if hasattr(subnet,"gateway"):
                                assert ip_address != parse_ip4_address(subnet.gateway)
                        except Exception:
                            raise ValueError("Specified ip_address " + parameters['ip_address'] + " is not an usable IP address in subnet " + subnet.id)

                        if str(ip_address) not in subnet_map.allocated_ips:
                            subnet_map.allocated_ips[str(ip_address)] = value.create_weakreference()
                            write(subnet_map.getkey(), subnet_map)
                        else:
                            raise ValueError("Cannot allocate an available IP address from subnet " + subnet.id)
                for k, v in parameters.items():
                    if k not in ('id',):
                        setattr(value, k, v)
                write(key, value)

        await call_api(self.app_routine, 'objectdb', 'writewalk', {'keys': tuple(parameter_dict.keys()),
                                                                   'walker': walker})
        return await self._dumpkeys(parameter_dict)

    async def deletelogicalport(self, id: str):
        "Delete logical port"
        p = {"id":id}
        return await self.deletelogicalports([p])
    
    @checked
    async def deletelogicalports(self, ports: [{"id": str}]):
        "Delete multiple logical ports"
        parameter_dict = OrderedDict()
        for port in ports:
            key = LogicalPort.default_key(port['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: " + port['id'])
            parameter_dict[key] = port
        def walker(walk, write):
            for key, parameters in parameter_dict.items():
                with suppress(WalkKeyNotRetrieved):
                    value = walk(key)
                    if value is None:
                        raise ValueError("Logical port " + parameters['id'] + " not exists")
                    with suppress(WalkKeyNotRetrieved):
                        lognet_map = walk(LogicalNetworkMap._network.leftkey(value.network))
                        lognet_map.ports.dataset().discard(value.create_weakreference())
                        write(lognet_map.getkey(), lognet_map)
                    if hasattr(value, 'subnet'):
                        with suppress(WalkKeyNotRetrieved):
                            subnet_map = walk(SubNetMap._subnet.leftkey(value.subnet))
                            del subnet_map.allocated_ips[str(parse_ip4_address(value.ip_address))]
                            write(subnet_map.getkey(), subnet_map)
                    with suppress(WalkKeyNotRetrieved):
                        logport_set = walk(LogicalPortSet.default_key())
                        logport_set.set.dataset().discard(value.create_weakreference())
                        write(logport_set.getkey(), logport_set)
                    write(key, None)
        await call_api(self.app_routine, 'objectdb', 'writewalk',{'keys': tuple(parameter_dict) + (LogicalPortSet.default_key(),),
                                                                  'walker': walker})
        return {"status":'OK'}

    async def listlogicalports(self,id = None,logicalnetwork = None,**kwargs):
        """
        Query logical port
        
        :param id: If specified, returns only logical port with this ID.
        
        :param logicalnetwork: If specified, returns only logical ports in this network.
        
        :param \*\*kwargs: customzied filters
        
        :return: return matched logical ports
        """
        if id:
            # specify id ,  find it ,, filter it
            lgportkey = LogicalPort.default_key(id)
            return await self._dumpone(lgportkey,kwargs)
        else:
            if logicalnetwork:
                # specify logicalnetwork , find in logicalnetwork map , filter it
                lgnet_map_key = LogicalNetworkMap.default_key(logicalnetwork)

                self._reqid += 1
                reqid = ('viperflow',self._reqid)
                def walk_map(key,value,walk,save):
                    if value is None:
                        return

                    for weakobj in value.ports.dataset():
                        lgportkey = weakobj.getkey()
                        with suppress(WalkKeyNotRetrieved):
                            lgport_obj = walk(lgportkey)
                            if all(getattr(lgport_obj,k,None) == v for k,v in kwargs.items()):
                                save(lgportkey)
                with request_context(reqid, self.app_routine):
                    _, values = await call_api(self.app_routine,'objectdb','walk',
                                               {'keys':[lgnet_map_key],
                                                'walkerdict':{lgnet_map_key:walk_map},
                                                'requestid':reqid})
                    return [dump(r) for r in values]

            else:
                logport_set_key = LogicalPortSet.default_key()

                self._reqid += 1
                reqid = ('viperflow',self._reqid)

                def walk_set(key,value,walk,save):
                    if value is None:
                        return

                    for weakobj in value.set.dataset():
                        lgportkey = weakobj.getkey()
                        try:
                            lgport_obj = walk(lgportkey)
                        except KeyError:
                            pass
                        else:
                            if all(getattr(lgport_obj,k,None) == v for k,v in kwargs.items()):
                                save(lgportkey)
                with request_context(reqid, self.app_routine):
                    _, values = await call_api(self.app_routine,"objectdb","walk",{'keys':[logport_set_key],
                                                                "walkerdict":{logport_set_key:walk_set},
                                                                "requestid":reqid})
                    return [dump(r) for r in values]

    async def createsubnet(self, logicalnetwork: str,
                                 cidr: cidr_type,
                                 id: (str, None) = None,
                                 **kwargs: {"?gateway": ip_address_type,
                                            "?allocated_start": ip_address_type,
                                            "?allocated_end": ip_address_type,
                                            "?host_routes": [tuple_((cidr_nonstrict_type, ip_address_type))],
                                            "?isexternal": bool,
                                            "?pre_host_config": [{"?systemid": str,
                                                                  "?bridge": str,
                                                                  "?vhost": str,
                                                                  "?cidr": cidr_type,
                                                                  "?local_address": ip_address_type,
                                                                  "?gateway": ip_address_type}],
                                            "?mtu": autoint,
                                            "?dns_nameservers": ([ip_address_type], None),
                                            "?domain_name": (str, None),
                                            "?ntp_servers": ([ip_address_type], None),
                                            "?lease_time": lease_time_type,
                                            "?extra_dhcp_options": dhcp_options_type}):
        """
        Create a subnet for the logical network.
        
        :param logicalnetwork: The logical network is subnet is in.
        
        :param cidr: CIDR of this subnet like ``"10.0.1.0/24"``
        
        :param id: subnet ID. If omitted, an UUID is generated.
        
        :param \*\*kwargs: customized creating options. Common options are:
                           
                           gateway
                              Gateway address for this subnet
                           
                           allocated_start
                              First IP of the allowed IP range.
                           
                           allocated_end
                              Last IP of the allowed IP range.
                           
                           host_routes
                              A list of ``[dest_cidr, via]`` like
                              ``[["192.168.1.0/24", "192.168.2.3"],["192.168.3.0/24","192.168.2.4"]]``.
                              This creates static routes on the subnet.

                           isexternal
                              This subnet can forward packet to external physical network

                           pre_host_config
                              A list of ``[{systemid, bridge, cidr, local_address, gateway, ...}]``
                              Per host configuration, will union with public info when used
        
        :return: A dictionary of information of the subnet.
        """
        if not id:
            id = str(uuid1())
        subnet = {'id':id,'logicalnetwork':logicalnetwork,'cidr':cidr}
        subnet.update(kwargs)

        return await self.createsubnets([subnet])
    
    @checked
    async def createsubnets(self, subnets: [{"?id": str,
                                             "logicalnetwork": str,
                                             "cidr": cidr_type,
                                             "?gateway": ip_address_type,
                                             "?allocated_start": ip_address_type,
                                             "?allocated_end": ip_address_type,
                                             "?host_routes": [tuple_((cidr_nonstrict_type, ip_address_type))],
                                             "?isexternal": bool,
                                             "?pre_host_config": [{"?systemid": str,
                                                                   "?bridge": str,
                                                                   "?vhost": str,
                                                                   "?cidr": cidr_type,
                                                                   "?local_address": ip_address_type,
                                                                   "?gateway": ip_address_type}],
                                             "?mtu": autoint,
                                             "?dns_nameservers": ([ip_address_type], None),
                                             "?domain_name": (str, None),
                                             "?ntp_servers": ([ip_address_type], None),
                                             "?lease_time": lease_time_type,
                                             "?extra_dhcp_options": dhcp_options_type}]):
        """
        Create multiple subnets in a transaction.
        """
        parameter_dict = OrderedDict()
        for subnet in subnets:
            subnet = copy.deepcopy(subnet)
            if 'id' not in subnet:
                subnet['id'] = str(uuid1())
            key = SubNet.default_key(subnet['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: " + subnet['id'])
            cidr, prefix = parse_ip4_network(subnet['cidr'])
            if 'gateway' in subnet:
                gateway = parse_ip4_address(subnet['gateway'])
                if not ip_in_network(gateway, cidr, prefix):
                    raise ValueError(" %s not in cidr %s" % (subnet['gateway'], subnet["cidr"]))
            if 'router' in subnet:
                raise ValueError("Cannot directly update router; use virtual router APIs")
            if "allocated_start" not in subnet:
                start = network_first(cidr, prefix)
            else:
                start = parse_ip4_address(subnet['allocated_start'])
                if not ip_in_network(start, cidr, prefix):
                    raise ValueError("%s not in cidr %s" % (subnet['allocated_start'], subnet["cidr"]))
        
            if "allocated_end" not in subnet:
                end = network_last(cidr,prefix)
            else:
                end = parse_ip4_address(subnet['allocated_end'])
                if not ip_in_network(end, cidr, prefix):
                    raise ValueError("%s not in cidr %s" % (subnet['allocated_end'], subnet["cidr"]))
            if start > end:
                raise ValueError("IP pool is empty for subnet " + subnet['id'])
            subnet['allocated_start'] = ip4_addr.formatter(start)
            subnet['allocated_end'] = ip4_addr.formatter(end)
            if 'pre_host_config' in subnet:
                for d in subnet['pre_host_config']:
                    if 'systemid' not in d:
                        d['systemid'] = "%"
                    if 'bridge' not in d:
                        d['bridge'] = '%'
                    if 'vhost' not in d:
                        d['vhost'] = ''

                    if 'cidr' in d:
                        local_cidr, local_prefix = parse_ip4_network(d['cidr'])
                        local_cidr_str = d['cidr']
                    else:
                        local_cidr, local_prefix = cidr, prefix
                        local_cidr_str = subnet['cidr']
                    if 'local_address' in d:
                        local = parse_ip4_address(d['local_address'])
                        if not ip_in_network(local, local_cidr, local_prefix):
                            raise ValueError(" %s not in cidr %s" % (d['local_address'], local_cidr_str))
                    if 'gateway' in d:
                        gateway = parse_ip4_address(d['gateway'])
                        if not ip_in_network(gateway, local_cidr, local_prefix):
                            raise ValueError(" %s not in cidr %s" % (d['gateway'], local_cidr_str))
            parameter_dict[key] = subnet
        
        keys = set(parameter_dict)
        keys.update(SubNetMap.default_key(subnet['id']) for subnet in parameter_dict.values())
        keys.update(LogicalNetworkMap.default_key(subnet['logicalnetwork']) for subnet in parameter_dict.values())
        keys.add(SubNetSet.default_key())
        def walker(walk, write):
            for key, parameters in parameter_dict.items():
                with suppress(WalkKeyNotRetrieved):
                    value = walk(key)
                    value = set_new(value, SubNet.create_instance(parameters['id']))
                    subnet_map = SubNetMap.create_instance(parameters['id'])
                    with suppress(WalkKeyNotRetrieved):
                        network_map = walk(LogicalNetworkMap.default_key(parameters['logicalnetwork']))
                        if network_map is None:
                            raise ValueError("Logical network " + parameters['logicalnetwork'] + " not exists")
                        value.network = ReferenceObject(LogicalNetwork.default_key(parameters['logicalnetwork']))
                        network_map.subnets.dataset().add(value.create_weakreference())
                        write(network_map.getkey(), network_map)
                    for k,v in parameters.items():
                        if k not in ('id', 'logicalnetwork', 'network'):
                            setattr(value, k, v)
                    with suppress(WalkKeyNotRetrieved):
                        subnet_set = walk(SubNetSet.default_key())
                        subnet_set.set.dataset().add(value.create_weakreference())
                        write(subnet_set.getkey(), subnet_set)
                    write(key, value)
                    write(subnet_map.getkey(), subnet_map)
        await call_api(self.app_routine,'objectdb','writewalk',{"keys":tuple(keys),
                                                                'walker':walker})
        return await self._dumpkeys(parameter_dict)

    async def updatesubnet(self,id: str, **kwargs: {"?cidr": cidr_type,
                                                    "?gateway": ip_address_type,
                                                    "?allocated_start": ip_address_type,
                                                    "?allocated_end": ip_address_type,
                                                    "?host_routes": [tuple_((cidr_nonstrict_type, ip_address_type))],
                                                    "?isexternal": bool,
                                                    "?pre_host_config": [{"?systemid": str,
                                                                          "?bridge": str,
                                                                          "?vhost": str,
                                                                          "?cidr": cidr_type,
                                                                          "?local_address": ip_address_type,
                                                                          "?gateway": ip_address_type}],
                                                    "?mtu": autoint,
                                                    "?dns_nameservers": ([ip_address_type], None),
                                                    "?domain_name": (str, None),
                                                    "?ntp_servers": ([ip_address_type], None),
                                                    "?lease_time": lease_time_type,
                                                    "?extra_dhcp_options": dhcp_options_type}):
        """
        Update subnet attributes
        """
        subnet = {"id":id}
        subnet.update(kwargs)
        return await self.updatesubnets([subnet])
    
    @checked
    async def updatesubnets(self, subnets: [{"id": str,
                                             "?cidr": cidr_type,
                                             "?gateway": ip_address_type,
                                             "?allocated_start": ip_address_type,
                                             "?allocated_end": ip_address_type,
                                             "?host_routes": [tuple_((cidr_nonstrict_type, ip_address_type))],
                                             "?pre_host_config": [{"?systemid": str,
                                                                   "?bridge": str,
                                                                   "?vhost": str,
                                                                   "?cidr": cidr_type,
                                                                   "?local_address": ip_address_type,
                                                                   "?gateway": ip_address_type}],
                                             "?mtu": autoint,
                                             "?dns_nameservers": ([ip_address_type], None),
                                             "?domain_name": (str, None),
                                             "?ntp_servers": ([ip_address_type], None),
                                             "?lease_time": lease_time_type,
                                             "?extra_dhcp_options": dhcp_options_type}]):
        """
        Update multiple subnets
        """
        parameter_dict = OrderedDict()
        for subnet in subnets:
            subnet = copy.deepcopy(subnet)
            if 'logicalnetwork' in subnet:
                raise ValueError("logical network cannot be changed")
            if "isexternal" in subnet:
                raise ValueError("isexternal cannot be changed")
            if 'router' in subnet:
                raise ValueError("Cannot directly update router; use virtual router APIs")
            key = SubNet.default_key(subnet['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: " + subnet['id'])
            if 'pre_host_config' in subnet:
                for d in subnet['pre_host_config']:
                    if 'systemid' not in d:
                        d['systemid'] = "%"
                    if 'bridge' not in d:
                        d['bridge'] = '%'
                    if 'vhost' not in d:
                        d['vhost'] = ''
                    if 'cidr' in d:
                        d['cidr'] = format_network_cidr(d['cidr'], True)
                        local_cidr, local_prefix = parse_ip4_network(d['cidr'])
                        if 'local_address' in d:
                            local = parse_ip4_address(d['local_address'])
                            if not ip_in_network(local, local_cidr, local_prefix):
                                raise ValueError(" %s not in cidr %s" % (d['local_address'], d['cidr']))
                        if 'gateway' in d:
                            gateway = parse_ip4_address(d['gateway'])
                            if not ip_in_network(gateway, local_cidr, local_prefix):
                                raise ValueError(" %s not in cidr %s" % (d['gateway'], d['cidr']))
            parameter_dict[key] = subnet

        subnetkeys = [SubNet.default_key(sn['id']) for sn in subnets]
        subnetmapkeys = [SubNetMap.default_key(sn['id']) for sn in subnets]

        keys = itertools.chain(subnetkeys,subnetmapkeys)
        
        def walker(walk, write):
            for key, parameters in parameter_dict.items():
                with suppress(WalkKeyNotRetrieved):
                    value = walk(key)
                    if value is None:
                        raise ValueError("Subnet " + parameters['id'] + " not exists")
                    subnet_map = walk(SubNetMap._subnet.leftkey(key))
                    for k, v in parameters.items():
                        if k not in ('id', 'network', 'logicalnetwork'):
                            setattr(value, k, v)
                    # Check consistency
                    try:
                        check_ip_pool(gateway=getattr(value, 'gateway', None),
                                      start=value.allocated_start,
                                      end=value.allocated_end,
                                      allocated=subnet_map.allocated_ips,
                                      cidr=value.cidr)
                    except Exception:
                        raise ValueError("New configurations conflicts with the old configuration or allocations in subnet "\
                                        + parameters['id'])
                    # Check pre_host_config
                    if hasattr(value, 'pre_host_config'):
                        cidr, prefix = parse_ip4_network(value.cidr)
                        for d in value.pre_host_config:
                            if 'cidr' in d:
                                local_cidr, local_prefix = parse_ip4_network(d['cidr'])
                                local_cidr_str = d['cidr']
                            else:
                                local_cidr, local_prefix = cidr, prefix
                                local_cidr_str = value.cidr
                            if 'local_address' in d:
                                local = parse_ip4_address(d['local_address'])
                                if not ip_in_network(local, local_cidr, local_prefix):
                                    raise ValueError(" %s not in cidr %s" % (d['local_address'], local_cidr_str))
                            if 'gateway' in d:
                                gateway = parse_ip4_address(d['gateway'])
                                if not ip_in_network(gateway, local_cidr, local_prefix):
                                    raise ValueError(" %s not in cidr %s" % (d['gateway'], local_cidr_str))
                    write(key, value)
        await call_api(self.app_routine, 'objectdb', 'writewalk', {'keys':keys,'walker':walker})
        return await self._dumpkeys(parameter_dict)

    async def deletesubnet(self, id: str):
        """
        Delete subnet
        """
        subnet = {"id":id}

        return await self.deletesubnets([subnet])

    @checked
    async def deletesubnets(self, subnets: [{"id": str}]):
        """
        Delete multiple subnets
        """
        parameter_dict = OrderedDict()
        for subnet in subnets:
            key = SubNet.default_key(subnet['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: " + subnet['id'])
            parameter_dict[key] = subnet

        subnetkeys = [SubNet.default_key(sn['id']) for sn in subnets]
        subnetmapkeys = [SubNetMap.default_key(sn['id']) for sn in subnets]

        keys = itertools.chain(subnetkeys,subnetmapkeys,[SubNetSet.default_key()])
        def walker(walk, write):
            for key, parameters in parameter_dict.items():
                with suppress(WalkKeyNotRetrieved):
                    value = walk(key)
                    if value is None:
                        raise ValueError("Subnet " + parameters['id'] + " not exists")
                    if hasattr(value, 'router'):
                        raise ValueError("Subnet " + parameters['id'] + " is in router " + value.router.getkey())
                    subnet_map = walk(SubNetMap._subnet.leftkey(key))
                    if subnet_map.allocated_ips:
                        raise ValueError("Subnet " + parameters['id'] + " has logical ports or router ports")
                    if hasattr(subnet_map, 'routers') and subnet_map.routers.dataset():
                        raise ValueError("Subnet " + parameters['id'] + " has router ports")
                    with suppress(WalkKeyNotRetrieved):
                        # Remove from logical network map
                        lognet_map = walk(LogicalNetworkMap._network.leftkey(value.network))
                        lognet_map.subnets.dataset().discard(value.create_weakreference())
                        write(lognet_map.getkey(), lognet_map)
                    with suppress(WalkKeyNotRetrieved):
                        subnet_set = walk(SubNetSet.default_key())
                        subnet_set.set.dataset().discard(value.create_weakreference())
                        write(subnet_set.getkey(), subnet_set)
                    write(key, None)
                    write(subnet_map.getkey(), None)
        await call_api(self.app_routine, "objectdb", "writewalk", {'keys': keys, 'walker': walker})
        return {"status":'OK'}

    async def listsubnets(self,id = None,logicalnetwork=None,**kwargs):
        """
        Query subnets
        
        :param id: if specified, only return subnet with this ID
        
        :param logicalnetwork: if specified, only return subnet in the network
        
        :param \*\*kwargs: customized filters
        
        :return: A list of dictionaries each stands for a matched subnet.
        """
        if id:
            # specify id , find it ,, filter it
            subnet_key = SubNet.default_key(id)

            return await self._dumpone(subnet_key,kwargs)

        else:
            if logicalnetwork:
                # specify logicalnetwork ,, find in logicalnetwork ,, filter it

                def walk_map(key,value,walk,save):
                    if value is None:
                        return

                    for weakobj in value.subnets.dataset():
                        subnet_key = weakobj.getkey()
                        with suppress(WalkKeyNotRetrieved):
                            subnet_obj = walk(subnet_key)
                            if all(getattr(subnet_obj,k,None) == v for k,v in kwargs.items()):
                                save(subnet_key)

                lgnetmap_key = LogicalNetworkMap.default_key(logicalnetwork)

                self._reqid += 1
                reqid = ("viperflow",self._reqid)
                with request_context(reqid, self.app_routine):
                    _, values = await call_api(self.app_routine,"objectdb","walk",
                                                {'keys':[lgnetmap_key],
                                                 'walkerdict':{lgnetmap_key:walk_map},
                                                 "requestid":reqid})
                    return [dump(r) for r in values]
            else:
                # find in all set ,, filter it
                subnet_set_key = SubNetSet.default_key()

                self._reqid += 1
                reqid = ('viperflow',self._reqid)

                def walk_set(key,value,walk,save):
                    if value is None:
                        return

                    for weakobj in value.set.dataset():
                        subnet_key = weakobj.getkey()

                        try:
                            subnet_obj = walk(subnet_key)
                        except KeyError:
                            pass
                        else:
                            if all(getattr(subnet_obj,k,None) == v for k,v in kwargs.items()):
                                save(subnet_key)
                with request_context(reqid, self.app_routine):
                    _, values = await call_api(self.app_routine,"objectdb","walk",
                                              {'keys':[subnet_set_key],
                                               'walkerdict':{subnet_set_key:walk_set},
                                               'requestid':reqid})
                    return [dump(r) for r in values]

    # the first run as routine going
    async def load(self,container):

        # init callback set dataobject as an transaction
        # args [values] order as keys
        @updater
        def init(physet,phyportset,logicset,logicportset,subnetset):
            if physet is None:
                physet = PhysicalNetworkSet()
            if phyportset is None:
                phyportset = PhysicalPortSet()
            if logicset is None:
                logicset = LogicalNetworkSet()
            if logicportset is None:
                logicportset = LogicalPortSet()
            if subnetset is None:
                subnetset = SubNetSet()

            return [physet,phyportset,logicset,logicportset,subnetset]

        # dataobject keys that will be init ,, add it if necessary
        initdataobjectkeys = [PhysicalNetworkSet.default_key(),
                PhysicalPortSet.default_key(),LogicalNetworkSet.default_key(),
                LogicalPortSet.default_key(),SubNetSet.default_key()]
        await call_api(container,'objectdb','transact',
                       {'keys':initdataobjectkeys,'updater':init})

        # call so main routine will be run
        return await Module.load(self,container)


def check_ip_pool(gateway, start, end, allocated, cidr):

    nstart = parse_ip4_address(start)
    nend = parse_ip4_address(end)
    ncidr,prefix = parse_ip4_network(cidr)
    if gateway:
        ngateway = parse_ip4_address(gateway)
        assert ip_in_network(nstart,ncidr,prefix)
        assert ip_in_network(nend,ncidr,prefix)
        assert nstart <= nend
        for ip, target in allocated.items():
            nip = parse_ip4_address(ip)
            assert ip_in_network(nip, ncidr, prefix)
            if isinstance(target, WeakReferenceObject):
                assert ip != ngateway
                assert nstart <= nip <= nend
    else:
        assert ip_in_network(nstart,ncidr,prefix)
        assert ip_in_network(nend,ncidr,prefix)
        assert nstart <= nend

        for ip, target in allocated.items():
            nip = parse_ip4_address(ip)
            assert ip_in_network(nip, ncidr, prefix)
            if isinstance(target, WeakReferenceObject):
                assert nstart <= nip <= nend




'''
Created on 2016/3/24

:author: hubo
'''
from vlcp.config.config import defaultconfig
import vlcp.service.kvdb.storage as storage
import vlcp.service.kvdb.redisnotifier as redisnotifier
from vlcp.server.module import depend, Module, call_api, api
import vlcp.utils.dataobject as dataobj
from vlcp.event.runnable import RoutineContainer
from vlcp.event.event import Event, withIndices, M_
from time import time
from copy import deepcopy
from vlcp.event.core import QuitException, syscall_removequeue
import itertools
from vlcp.utils.dataobject import AlreadyExistsException, UniqueKeyReference,\
    MultiKeyReference, DataObjectSet, UniqueKeySet, WeakReferenceObject,\
    MultiKeySet, ReferenceObject, request_context
from contextlib import closing
import functools
import copy
from vlcp.utils.exceptions import AsyncTransactionLockException, StaleResultException,\
    TransactionRetryExceededException, TransactionTimeoutException, WalkKeyNotRetrieved

try:
    from itertools import izip
except ImportError:
    izip = zip

@withIndices()
class RetrieveRequestSend(Event):
    pass

@withIndices('id')
class RetrieveReply(Event):
    pass

def _str(b):
    if isinstance(b, str):
        return b
    elif isinstance(b, bytes):
        return b.decode('utf-8')
    else:
        return str(b)

def _str2(b):
    if isinstance(b, str):
        return b
    elif isinstance(b, bytes):
        return b.decode('utf-8')
    elif hasattr(b, 'getkey'):
        return b.getkey()
    else:
        return str(b)


class _NeedMoreKeysException(Exception):
    pass


@defaultconfig
@depend(storage.KVStorage, redisnotifier.UpdateNotifier)
class ObjectDB(Module):
    """
    Abstract transaction layer for KVDB
    """
    service = True
    # Priority for object update event
    _default_objectupdatepriority = 450
    # Enable debugging mode for updater: all updaters will be called for an extra time
    # to make sure it does not crash with multiple calls
    _default_debuggingupdater = False
    def __init__(self, server):
        Module.__init__(self, server)
        self._managed_objs = {}
        self._watches = {}
        self._requestids = {}
        self._watchedkeys = set()
        self._requests = []
        self._transactno = 0
        self._stale = False
        self._updatekeys = set()
        self._update_version = {}
        self._cache = None
        self._pending_gc = 0
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._update
        self.routines.append(self.apiroutine)
        self.create_api(api(self.mget, self.apiroutine),
                       api(self.get, self.apiroutine),
                       api(self.mgetonce, self.apiroutine),
                       api(self.getonce, self.apiroutine),
                       api(self.mwatch, self.apiroutine),
                       api(self.watch, self.apiroutine),
                       api(self.munwatch, self.apiroutine),
                       api(self.unwatch, self.apiroutine),
                       api(self.unwatchall, self.apiroutine),
                       api(self.transact, self.apiroutine),
                       api(self.watchlist),
                       api(self.walk, self.apiroutine),
                       api(self.gettimestamp, self.apiroutine),
                       api(self.asynctransact, self.apiroutine),
                       api(self.writewalk, self.apiroutine),
                       api(self.asyncwritewalk, self.apiroutine)
                       )

    def _set_watch(self, key, requestid):
        self._watches.setdefault(key, set()).add(requestid)
        self._requestids.setdefault(requestid, set()).add(key)

    def _remove_watch(self, key, requestid):
        s = self._watches.get(key)
        if s:
            s.discard(requestid)
            if not s:
                del self._watches[key]
        s = self._requestids.get(requestid)
        if s:
            s.discard(key)
            if not s:
                del self._requestids[requestid]
    
    def _remove_all_watches(self, requestid):
        s = self._requestids.get(requestid)
        if s is not None:
            for k in s:
                s2 = self._watches.get(k)
                if s2:
                    s2.discard(requestid)
                    if not s2:
                        del self._watches[k]
            del self._requestids[requestid]

    async def load(self, container):
        self.scheduler.queue.addSubQueue(\
                self.objectupdatepriority, dataobj.DataObjectUpdateEvent.createMatcher(), 'dataobjectupdate')
        self._notifier = await call_api(container, 'updatenotifier', 'createnotifier')
        await Module.load(self, container)
        self.routines.append(self._notifier)

    async def unload(self, container, force=False):
        await container.syscall(syscall_removequeue(self.scheduler.queue, 'dataobjectupdate'))
        await Module.unload(self, container, force=force)

    async def _update(self):
        timestamp = '%012x' % (int(time() * 1000),) + '-'
        notification_matcher = self._notifier.notification_matcher(False)
        def copywithkey(obj, key):
            newobj = deepcopy(obj)
            if hasattr(newobj, 'setkey'):
                newobj.setkey(key)
            return newobj
        def getversion(obj):
            if obj is None:
                return (0, -1)
            else:
                return (getattr(obj, 'kvdb_createtime', 0), getattr(obj, 'kvdb_updateversion', 0))
        def isnewer(obj, version):
            if obj is None:
                return version[1] != -1
            else:
                return getversion(obj) > version
        request_matcher = RetrieveRequestSend.createMatcher()
        def onupdate(event, matcher):
            update_keys = self._watchedkeys.intersection([_str(k) for k in event.keys])
            self._updatekeys.update(update_keys)
            if event.extrainfo:
                for k,v in zip(event.keys, event.extrainfo):
                    k = _str(k)
                    if k in update_keys:
                        v = tuple(v)
                        oldv = self._update_version.get(k, (0, -1))
                        if oldv < v:
                            self._update_version[k] = v
            else:
                for k in event.keys:
                    try:
                        del self._update_version[_str(k)]
                    except KeyError:
                        pass
        async def updateinner():
            processing_requests = []
            # New managed keys
            retrieve_list = set()
            orig_retrieve_list = set()
            retrieveonce_list = set()
            orig_retrieveonce_list = set()
            processing_request_ids = set()
            # Retrieved values are stored in update_result before merging into current storage
            update_result = {}
            # key => [(walker_func, (original_keys, rid)), ...]
            walkers = {}
            # Use the loop count as a revision identifier, then the valid revisions of the value
            # in update_result is a range, from the last loop count the value changed
            # (or -1 if not changed), to the last loop count the value is retrieved
            #
            # each walker can only walk on keys that shares at least one revision to ensure the
            # values are consistent. If no revision could be shared, all the keys must be retrieved
            # again to get a consistent view
            revision_min = {}
            revision_max = {}
            self._loopCount = 0
            # A request-id -> retrieve set dictionary to store the saved keys
            savelist = {}
            
            # (start_key, walker_func, rid) => set(used_keys)
            walker_used_keys = {}
            
            # used_key => [(start_key, walker_func, (original_keys, rid)), ...]
            used_key_ref = {}
            
            def _update_walker_ref(start_key, walker, original_keys, rid, used_keys):
                old_used_keys = walker_used_keys.get((start_key, walker, rid), ())
                for k in old_used_keys:
                    if k not in used_keys:
                        old_list = used_key_ref[k]
                        for i, v in enumerate(old_list):
                            if v[0] == start_key and v[1] == walker and v[2][1] == rid:
                                break
                        else:
                            continue
                        old_list[i:] = old_list[i+1:]
                for k in used_keys:
                    if k not in old_used_keys:
                        used_key_ref.setdefault(k, []).append((start_key, walker, (original_keys, rid)))
                walker_used_keys[(start_key, walker, rid)] = set(used_keys)
            
            # (start_key, walker, rid) => cached_result
            finished_walkers = {}
            
            def _dirty_walkers(new_values):
                for k in new_values:
                    if k in used_key_ref:
                        for start_key, walker, (_, rid) in used_key_ref[k]:
                            finished_walkers.pop((start_key, walker, rid), None)
            
            async def updateloop():
                while (retrieve_list or self._updatekeys or self._requests):
                    # default walker, default walker cached, customized walker, customized walker cached
                    _performance_counters = [0, 0, 0, 0]
                    # Updated keys
                    update_list = set()
                    if self._loopCount >= 10 and not retrieve_list:
                        if not self._updatekeys:
                            break
                        elif self._loopCount >= 100:
                            # Too many updates, we must stop to respond
                            self._logger.warning("There are still database updates after 100 loops of mget, respond with potential inconsistent values")
                            break
                    if self._updatekeys:
                        update_list.update(self._updatekeys)
                        self._updatekeys.clear()
                    if self._requests:
                        # Processing requests
                        for r in self._requests:
                            if r[2] == 'unwatch':
                                try:
                                    for k in r[0]:
                                        self._remove_watch(k, r[3])
                                    # Do not need to wait
                                except Exception as exc:
                                    await self.apiroutine.wait_for_send(RetrieveReply(r[1], exception = exc))
                                else:
                                    await self.apiroutine.wait_for_send(RetrieveReply(r[1], result = None))
                            elif r[2] == 'unwatchall':
                                if r[3] in processing_request_ids:
                                    # unwatch a processing request
                                    # pend this request until all requests are processed
                                    processing_requests.append(r)
                                else:
                                    try:
                                        self._remove_all_watches(r[3])
                                    except Exception as exc:
                                        await self.apiroutine.wait_for_send(RetrieveReply(r[1], exception = exc))
                                    else:
                                        await self.apiroutine.wait_for_send(RetrieveReply(r[1], result = None))
                            elif r[2] == 'watch':
                                retrieve_list.update(r[0])
                                orig_retrieve_list.update(r[0])
                                for k in r[0]:
                                    self._set_watch(k, r[3])
                                processing_requests.append(r)
                                processing_request_ids.add(r[3])
                            elif r[2] == 'get':
                                retrieve_list.update(r[0])
                                orig_retrieve_list.update(r[0])
                                processing_requests.append(r)
                                processing_request_ids.add(r[3])
                            elif r[2] == 'walk':
                                retrieve_list.update(r[0])
                                processing_requests.append(r)
                                for k,v in r[3].items():
                                    walkers.setdefault(k, []).append((v, (r[0], r[1])))
                                processing_request_ids.add(r[4])
                            else:
                                retrieveonce_list.update(r[0])
                                orig_retrieveonce_list.update(r[0])
                                processing_requests.append(r)
                        del self._requests[:]
                    if retrieve_list:
                        watch_keys = tuple(k for k in retrieve_list if k not in self._watchedkeys)
                        # Add watch_keys to notification
                        if watch_keys:
                            for k in watch_keys:
                                if k in update_result:
                                    self._update_version[k] = getversion(update_result[k])
                            await self._notifier.add_listen(*watch_keys)
                            self._watchedkeys.update(watch_keys)
                    get_list_set = update_list.union(itertools.chain((k for k in retrieve_list
                                                     if k not in update_result and k not in self._managed_objs),
                                                     (k for k in retrieveonce_list
                                                     if k not in update_result and k not in self._managed_objs)))
                    get_list = list(get_list_set)
                    new_values = set()
                    if get_list:
                        try:
                            result, self._cache = await call_api(
                                                            self.apiroutine,
                                                            'kvstorage',
                                                            'mgetwithcache',
                                                            {'keys': get_list, 'cache': self._cache}
                                                        )
                        except QuitException:
                            raise
                        except Exception:
                            # Serve with cache
                            if not self._stale:
                                self._logger.warning('KVStorage retrieve failed, serve with cache', exc_info = True)
                            self._stale = True
                            # Discard all retrieved results
                            update_result.clear()
                            # Retry update later
                            self._updatekeys.update(update_list)
                            #break
                            revision_min.clear()
                            revision_max.clear()
                        else:
                            self._stale = False
                            for k,v in izip(get_list, result):
                                # Update revision information
                                revision_max[k] = self._loopCount
                                if k not in update_result:
                                    if k not in self._managed_objs:
                                        # A newly retrieved key
                                        revision_min[k] = self._loopCount
                                        old_value = None
                                    else:
                                        old_value = self._managed_objs[k]
                                else:
                                    old_value = update_result[k]
                                # Check if the value is changed
                                if old_value is not v and getversion(old_value) != getversion(v):
                                    revision_min[k] = self._loopCount
                                    new_values.add(k)
                                else:
                                    if k not in revision_min:
                                        revision_min[k] = -1
                                if old_value is not v:
                                    if v is not None and hasattr(v, 'setkey'):
                                        v.setkey(k)
                                if k in self._watchedkeys and k not in self._update_version:
                                    self._update_version[k] = getversion(v)

                            update_result.update(zip(get_list, result))
                    # Disable cache for walkers with updated keys
                    _dirty_walkers(new_values)
                    
                    # All keys which should be retrieved in next loop
                    new_retrieve_list = set()
                    # Keys which should be retrieved in next loop for a single walk
                    new_retrieve_keys = set()
                    # Keys that are used in current walk will be retrieved again in next loop
                    used_keys = set()
                    # We separate the data with revisions to prevent inconsistent result
                    def create_walker(orig_key, strict=True):
                        revision_range = [revision_min.get(orig_key, -1), revision_max.get(orig_key, -1)]
                        def _walk_with_revision(key):
                            if hasattr(key, 'getkey'):
                                key = key.getkey()
                            key = _str(key)
                            if key not in self._watchedkeys:
                                # This key is not retrieved, raise a KeyError, and record this key
                                new_retrieve_keys.add(key)
                                raise WalkKeyNotRetrieved(key)
                            elif self._stale:
                                if key not in self._managed_objs:
                                    new_retrieve_keys.add(key)
                                used_keys.add(key)
                                return self._managed_objs.get(key)
                            elif key not in update_result and key not in self._managed_objs:
                                # This key is not retrieved, raise a KeyError, and record this key
                                new_retrieve_keys.add(key)
                                raise WalkKeyNotRetrieved(key)
                            # Check revision
                            current_revision = (
                                max(revision_min.get(key, -1), revision_range[0]),
                                min(revision_max.get(key, -1), revision_range[1])
                            )
                            if current_revision[1] < current_revision[0]:
                                # revisions cannot match
                                new_retrieve_keys.add(key)
                                if strict:
                                    used_keys.add(key)
                                    raise WalkKeyNotRetrieved(key)
                            else:
                                # update revision range
                                revision_range[:] = current_revision
                            if key in update_result:
                                used_keys.add(key)
                                return update_result[key]
                            else:
                                used_keys.add(key)
                                return self._managed_objs[key]
                        return _walk_with_revision
                    _default_walker_dup_check = set()
                    def default_walker(key, obj, walk, _circle_detect = None):
                        if _circle_detect is None:
                            _circle_detect = set()
                        if key in _circle_detect:
                            return
                        else:
                            _circle_detect.add(key)
                        if hasattr(obj, 'kvdb_internalref'):
                            rl = obj.kvdb_internalref()
                            for k in rl:
                                try:
                                    newobj = walk(k)
                                except KeyError:
                                    pass
                                else:
                                    if newobj is not None:
                                        default_walker(k, newobj, walk, _circle_detect)
                    def _do_default_walker(k):
                        if k not in _default_walker_dup_check:
                            _default_walker_dup_check.add(k)
                            _performance_counters[0] += 1
                            if (k, None, None) not in finished_walkers:
                                v = update_result.get(k)
                                if v is not None:
                                    new_retrieve_keys.clear()
                                    used_keys.clear()
                                    default_walker(k, v, create_walker(k, False))
                                    if new_retrieve_keys:
                                        new_retrieve_list.update(new_retrieve_keys)
                                        self._updatekeys.update(used_keys)
                                        self._updatekeys.add(k)
                                    else:
                                        _all_used_keys = used_keys.union([k])
                                        _update_walker_ref(k, None, None, None, _all_used_keys)
                                        finished_walkers[(k, None, None)] = None
                                else:
                                    _update_walker_ref(k, None, None, None, [k])
                                    finished_walkers[(k, None, None)] = None
                            else:
                                _performance_counters[1] += 1
                    for k in orig_retrieve_list:
                        _do_default_walker(k)
                    savelist.clear()
                    for k,ws in walkers.items():
                        # k: the walker key
                        # ws: list of [walker_func, (request_original_keys, rid)]
                        # Retry every walker, starts with k, with the value of v
                        if k in update_result:
                            # The value is newly retrieved
                            v = update_result.get(k)
                        else:
                            # Use the stored value
                            v = self._managed_objs.get(k)
                        if ws:
                            for w,r in list(ws):
                                # w: walker_func
                                # r: (request_original_keys, rid)
                                # Custom walker
                                _performance_counters[2] += 1
                                _cache_key = (k, w, r[1])
                                if _cache_key in finished_walkers:
                                    _performance_counters[3] += 1
                                    savelist.setdefault(r[1], set()).update(finished_walkers[_cache_key])
                                else:
                                    _local_save_list = set()
                                    def save(key):
                                        if hasattr(key, 'getkey'):
                                            key = key.getkey()
                                        key = _str(key)
                                        if key != k and key not in used_keys:
                                            raise ValueError('Cannot save a key without walk')
                                        _local_save_list.add(key)
                                    try:
                                        new_retrieve_keys.clear()
                                        used_keys.clear()
                                        w(k, v, create_walker(k), save)
                                    except Exception as exc:
                                        # if one walker failed, the whole request is failed, remove all walkers
                                        self._logger.warning("A walker raises an exception which rolls back the whole walk process. "
                                                             "walker = %r, start key = %r, new_retrieve_keys = %r, used_keys = %r",
                                                             w, k, new_retrieve_keys, used_keys, exc_info=True)
                                        for orig_k in r[0]:
                                            if orig_k in walkers:
                                                walkers[orig_k][:] = [(w0, r0) for w0,r0 in walkers[orig_k] if r0[1] != r[1]]
                                        processing_requests[:] = [r0 for r0 in processing_requests if r0[1] != r[1]]
                                        savelist.pop(r[1])
                                        await self.apiroutine.wait_for_send(RetrieveReply(r[1], exception = exc))
                                    else:
                                        savelist.setdefault(r[1], set()).update(_local_save_list)
                                        if new_retrieve_keys:
                                            new_retrieve_list.update(new_retrieve_keys)
                                            self._updatekeys.update(used_keys)
                                            self._updatekeys.add(k)
                                        else:
                                            _all_used_keys = used_keys.union([k])
                                            _update_walker_ref(k, w, r[0], r[1], _all_used_keys)
                                            finished_walkers[_cache_key] = _local_save_list
                    for save in savelist.values():
                        for k in save:
                            _do_default_walker(k)
                    retrieve_list.clear()
                    retrieveonce_list.clear()
                    retrieve_list.update(new_retrieve_list)
                    self._logger.debug("Loop %d: %d default walker (%d cached), %d customized walker (%d cached)",
                                       self._loopCount,
                                       *_performance_counters)
                    self._loopCount += 1
                    if self._stale:
                        watch_keys = tuple(k for k in retrieve_list if k not in self._watchedkeys)
                        if watch_keys:
                            await self._notifier.add_listen(*watch_keys)
                            self._watchedkeys.update(watch_keys)
                        break
            while True:
                await self.apiroutine.with_callback(updateloop(), onupdate, notification_matcher)
                if self._loopCount >= 100 or self._stale:
                    break
                # If some updated result is newer than the notification version, we should wait for the notification
                should_wait = False
                for k,v in update_result.items():
                    if k in self._watchedkeys:
                        oldv = self._update_version.get(k)
                        if oldv is not None and isnewer(v, oldv):
                            should_wait = True
                            break
                if should_wait:
                    timeout, ev, m = await self.apiroutine.wait_with_timeout(0.2, notification_matcher)
                    if timeout:
                        break
                    else:
                        onupdate(ev, m)
                else:
                    break
            # Update result
            send_events = []
            self._transactno += 1
            transactid = '%s%016x' % (timestamp, self._transactno)
            update_objs = []
            for k,v in update_result.items():
                if k in self._watchedkeys:
                    if v is None:
                        oldv = self._managed_objs.get(k)
                        if oldv is not None:
                            if hasattr(oldv, 'kvdb_detach'):
                                oldv.kvdb_detach()
                                update_objs.append((k, oldv, dataobj.DataObjectUpdateEvent.DELETED))
                            else:
                                update_objs.append((k, None, dataobj.DataObjectUpdateEvent.DELETED))
                            del self._managed_objs[k]
                    else:
                        oldv = self._managed_objs.get(k)
                        if oldv is not None:
                            if oldv != v:
                                if oldv and hasattr(oldv, 'kvdb_update'):
                                    oldv.kvdb_update(v)
                                    update_objs.append((k, oldv, dataobj.DataObjectUpdateEvent.UPDATED))
                                else:
                                    if hasattr(oldv, 'kvdb_detach'):
                                        oldv.kvdb_detach()
                                    self._managed_objs[k] = v
                                    update_objs.append((k, v, dataobj.DataObjectUpdateEvent.UPDATED))
                        else:
                            self._managed_objs[k] = v
                            update_objs.append((k, v, dataobj.DataObjectUpdateEvent.UPDATED))
            for k in update_result.keys():
                v = self._managed_objs.get(k)
                if v is not None and hasattr(v, 'kvdb_retrievefinished'):
                    v.kvdb_retrievefinished(self._managed_objs)
            allkeys = tuple(k for k,_,_ in update_objs)
            send_events.extend((dataobj.DataObjectUpdateEvent(k, transactid, t, object = v, allkeys = allkeys) for k,v,t in update_objs))
            # Process requests
            unwatchall = []
            for r in processing_requests:
                if r[2] == 'get':
                    objs = [self._managed_objs.get(k) for k in r[0]]
                    for k,v in zip(r[0], objs):
                        if v is not None:
                            self._set_watch(k, r[3])
                    result = [o.create_reference() if o is not None and hasattr(o, 'create_reference') else o
                              for o in objs]
                elif r[2] == 'watch':
                    result = [(v.create_reference() if hasattr(v, 'create_reference') else v)
                              if v is not None else dataobj.ReferenceObject(k)
                              for k,v in ((k,self._managed_objs.get(k)) for k in r[0])]
                elif r[2] == 'walk':
                    saved_keys = list(savelist.get(r[1], []))
                    for k in saved_keys:
                        self._set_watch(k, r[4])
                    objs = [self._managed_objs.get(k) for k in saved_keys]
                    result = (saved_keys,
                              [o.create_reference() if hasattr(o, 'create_reference') else o
                               if o is not None else dataobj.ReferenceObject(k)
                               for k,o in zip(saved_keys, objs)])
                elif r[2] == 'unwatchall':
                    # Remove watches after all results are processed
                    unwatchall.append(r[3])
                    result = None
                else:
                    result = [copywithkey(update_result.get(k, self._managed_objs.get(k)), k) for k in r[0]]
                send_events.append(RetrieveReply(r[1], result = result, stale = self._stale))
            for requestid in unwatchall:
                self._remove_all_watches(requestid)
            async def output_result():
                for e in send_events:
                    await self.apiroutine.wait_for_send(e)
            await self.apiroutine.with_callback(output_result(), onupdate, notification_matcher)
            self._pending_gc += 1
        async def _gc():
            # Use DFS to remove unwatched objects
            mark_set = set()
            def dfs(k):
                if k in mark_set:
                    return
                mark_set.add(k)
                v = self._managed_objs.get(k)
                if v is not None and hasattr(v, 'kvdb_internalref'):
                    for k2 in v.kvdb_internalref():
                        dfs(k2)
            for k in self._watches.keys():
                dfs(k)
            remove_keys = self._watchedkeys.difference(mark_set)
            if remove_keys:
                self._watchedkeys.difference_update(remove_keys)
                await self._notifier.remove_listen(*tuple(remove_keys))
                for k in remove_keys:
                    if k in self._managed_objs:
                        del self._managed_objs[k]
                    if k in self._update_version:
                        del self._update_version[k]
            if self._cache is not None:
                self._cache.gc(self._managed_objs)
            self._pending_gc = 0
        while True:
            if not self._updatekeys and not self._requests:
                if self._pending_gc >= 10:
                    await self.apiroutine.with_callback(_gc(), onupdate, notification_matcher)
                    continue
                elif self._pending_gc:
                    timeout, ev, m = await self.apiroutine.wait_with_timeout(1, notification_matcher, request_matcher)
                    if timeout:
                        await self.apiroutine.with_callback(_gc(), onupdate, notification_matcher)
                        continue
                else:
                    ev, m = await M_(notification_matcher, request_matcher)
                if m is notification_matcher:
                    onupdate(ev, m)
            await updateinner()

    async def mget(self, keys, requestid, nostale = False):
        "Get multiple objects and manage them. Return references to the objects."
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append((keys, rid, 'get', requestid))
        if notify:
            await self.apiroutine.wait_for_send(RetrieveRequestSend())
        ev = await RetrieveReply.createMatcher(rid)
        if hasattr(ev, 'exception'):
            raise ev.exception
        if nostale and ev.stale:
            raise StaleResultException(ev.result)
        return ev.result

    async def get(self, key, requestid, nostale = False):
        """
        Get an object from specified key, and manage the object.
        Return a reference to the object or None if not exists.
        """
        r = await self.mget([key], requestid, nostale)
        return r[0]

    async def mgetonce(self, keys, nostale = False):
        "Get multiple objects, return copies of them. Referenced objects are not retrieved."
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append((keys, rid, 'getonce'))
        if notify:
            await self.apiroutine.wait_for_send(RetrieveRequestSend())
        ev = await RetrieveReply.createMatcher(rid)
        if hasattr(ev, 'exception'):
            raise ev.exception
        if nostale and ev.stale:
            raise StaleResultException(ev.result)
        return ev.result

    async def getonce(self, key, nostale = False):
        "Get a object without manage it. Return a copy of the object, or None if not exists. Referenced objects are not retrieved."
        r = await self.mgetonce([key], nostale)
        return r[0]

    async def watch(self, key, requestid, nostale = False):
        """
        Try to find an object and return a reference. Use ``reference.isdeleted()`` to test
        whether the object exists.
        Use ``reference.wait(container)`` to wait for the object to be existed.
        """
        r = await self.mwatch([key], requestid, nostale)
        return r[0]

    async def mwatch(self, keys, requestid, nostale = False):
        "Try to return all the references, see ``watch()``"
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append((keys, rid, 'watch', requestid))
        if notify:
            await self.apiroutine.wait_for_send(RetrieveRequestSend())
        ev = await RetrieveReply.createMatcher(rid)
        if hasattr(ev, 'exception'):
            raise ev.exception
        if nostale and ev.stale:
            raise StaleResultException(ev.result)
        return ev.result

    async def unwatch(self, key, requestid):
        "Cancel management of a key"
        await self.munwatch([key], requestid)

    async def unwatchall(self, requestid):
        "Cancel management for all keys that are managed by requestid"
        notify = not self._requests
        rid = object()
        self._requests.append(((), rid, 'unwatchall', requestid))
        if notify:
            await self.apiroutine.wait_for_send(RetrieveRequestSend())
        ev = await RetrieveReply.createMatcher(rid)
        if hasattr(ev, 'exception'):
            raise ev.exception

    async def munwatch(self, keys, requestid):
        "Cancel management of keys"
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append((keys, rid, 'unwatch', requestid))
        if notify:
            await self.apiroutine.wait_for_send(RetrieveRequestSend())
        ev = await RetrieveReply.createMatcher(rid)
        if hasattr(ev, 'exception'):
            raise ev.exception

    async def transact(self, keys, updater, withtime = False, maxtime = 60):
        """
        Try to update keys in a transact, with an ``updater(keys, values)``,
        which returns ``(updated_keys, updated_values)``.
        
        The updater may be called more than once. If ``withtime = True``,
        the updater should take three parameters:
        ``(keys, values, timestamp)`` with timestamp as the server time
        """
        keys = tuple(_str2(k) for k in keys)
        updated_ref = [None, None]
        extra_keys = []
        extra_key_set = []
        auto_remove_keys = set()
        orig_len = len(keys)
        def updater_with_key(keys, values, timestamp):
            # Automatically manage extra keys
            remove_uniquekeys = []
            remove_multikeys = []
            update_uniquekeys = []
            update_multikeys = []
            keystart = orig_len + len(auto_remove_keys)
            for v in values[:keystart]:
                if v is not None:
                    if hasattr(v, 'kvdb_uniquekeys'):
                        remove_uniquekeys.extend((k,v.create_weakreference()) for k in v.kvdb_uniquekeys())
                    if hasattr(v, 'kvdb_multikeys'):
                        remove_multikeys.extend((k,v.create_weakreference()) for k in v.kvdb_multikeys())
            if self.debuggingupdater:
                # Updater may be called more than once, ensure that this updater does not crash
                # on multiple calls
                kc = keys[:orig_len]
                vc = [v.clone_instance() if v is not None and hasattr(v, 'clone_instance') else deepcopy(v) for v in values[:orig_len]]
                if withtime:
                    updated_keys, updated_values = updater(kc, vc, timestamp)
                else:
                    updated_keys, updated_values = updater(kc, vc)
            if withtime:
                updated_keys, updated_values = updater(keys[:orig_len], values[:orig_len], timestamp)
            else:
                updated_keys, updated_values = updater(keys[:orig_len], values[:orig_len])
            for v in updated_values:
                if v is not None:
                    if hasattr(v, 'kvdb_uniquekeys'):
                        update_uniquekeys.extend((k,v.create_weakreference()) for k in v.kvdb_uniquekeys())
                    if hasattr(v, 'kvdb_multikeys'):
                        update_multikeys.extend((k,v.create_weakreference()) for k in v.kvdb_multikeys())
            extrakeysdict = dict(zip(keys[keystart:keystart + len(extra_keys)], values[keystart:keystart + len(extra_keys)]))
            extrakeysetdict = dict(zip(keys[keystart + len(extra_keys):keystart + len(extra_keys) + len(extra_key_set)],
                                       values[keystart + len(extra_keys):keystart + len(extra_keys) + len(extra_key_set)]))
            tempdict = {}
            old_values = dict(zip(keys, values))
            updated_keyset = set(updated_keys)
            try:
                append_remove = set()
                autoremove_keys = set()
                # Use DFS to find auto remove keys
                def dfs(k):
                    if k in autoremove_keys:
                        return
                    autoremove_keys.add(k)
                    if k not in old_values:
                        append_remove.add(k)
                    else:
                        oldv = old_values[k]
                        if oldv is not None and hasattr(oldv, 'kvdb_autoremove'):
                            for k2 in oldv.kvdb_autoremove():
                                dfs(k2)
                for k,v in zip(updated_keys, updated_values):
                    if v is None:
                        dfs(k)
                if append_remove:
                    raise _NeedMoreKeysException()
                for k,v in remove_uniquekeys:
                    if v.getkey() not in updated_keyset and v.getkey() not in auto_remove_keys:
                        # This key is not updated, keep the indices untouched
                        continue
                    if k not in extrakeysdict:
                        raise _NeedMoreKeysException()
                    elif extrakeysdict[k] is not None and extrakeysdict[k].ref.getkey() == v.getkey():
                        # If the unique key does not reference to the correct object
                        # there may be an error, but we ignore this.
                        # Save in a temporary dictionary. We may restore it later.
                        tempdict[k] = extrakeysdict[k]
                        extrakeysdict[k] = None
                        setkey = UniqueKeyReference.get_keyset_from_key(k)
                        if setkey not in extrakeysetdict:
                            raise _NeedMoreKeysException()
                        else:
                            ks = extrakeysetdict[setkey]
                            if ks is None:
                                ks = UniqueKeySet.create_from_key(setkey)
                                extrakeysetdict[setkey] = ks
                            ks.set.dataset().discard(WeakReferenceObject(k))
                for k,v in remove_multikeys:
                    if v.getkey() not in updated_keyset and v.getkey() not in auto_remove_keys:
                        # This key is not updated, keep the indices untouched
                        continue
                    if k not in extrakeysdict:
                        raise _NeedMoreKeysException()
                    else:
                        mk = extrakeysdict[k]
                        if mk is not None:
                            mk.set.dataset().discard(v)
                            if not mk.set.dataset():
                                tempdict[k] = extrakeysdict[k]
                                extrakeysdict[k] = None
                                setkey = MultiKeyReference.get_keyset_from_key(k)
                                if setkey not in extrakeysetdict:
                                    raise _NeedMoreKeysException()
                                else:
                                    ks = extrakeysetdict[setkey]
                                    if ks is None:
                                        ks = MultiKeySet.create_from_key(setkey)
                                        extrakeysetdict[setkey] = ks
                                    ks.set.dataset().discard(WeakReferenceObject(k))
                for k,v in update_uniquekeys:
                    if k not in extrakeysdict:
                        raise _NeedMoreKeysException()
                    elif extrakeysdict[k] is not None and extrakeysdict[k].ref.getkey() != v.getkey():
                        raise AlreadyExistsException('Unique key conflict for %r and %r, with key %r' % \
                                                     (extrakeysdict[k].ref.getkey(), v.getkey(), k))
                    elif extrakeysdict[k] is None:
                        lv = tempdict.get(k, None)
                        if lv is not None and lv.ref.getkey() == v.getkey():
                            # Restore this value
                            nv = lv
                        else:
                            nv = UniqueKeyReference.create_from_key(k)
                            nv.ref = ReferenceObject(v.getkey())
                        extrakeysdict[k] = nv
                        setkey = UniqueKeyReference.get_keyset_from_key(k)
                        if setkey not in extrakeysetdict:
                            raise _NeedMoreKeysException()
                        else:
                            ks = extrakeysetdict[setkey]
                            if ks is None:
                                ks = UniqueKeySet.create_from_key(setkey)
                                extrakeysetdict[setkey] = ks
                            ks.set.dataset().add(nv.create_weakreference())
                for k,v in update_multikeys:
                    if k not in extrakeysdict:
                        raise _NeedMoreKeysException()
                    else:
                        mk = extrakeysdict[k]
                        if mk is None:
                            mk = tempdict.get(k, None)
                            if mk is None:
                                mk = MultiKeyReference.create_from_key(k)
                                mk.set = DataObjectSet()
                            setkey = MultiKeyReference.get_keyset_from_key(k)
                            if setkey not in extrakeysetdict:
                                raise _NeedMoreKeysException()
                            else:
                                ks = extrakeysetdict[setkey]
                                if ks is None:
                                    ks = MultiKeySet.create_from_key(setkey)
                                    extrakeysetdict[setkey] = ks
                                ks.set.dataset().add(mk.create_weakreference())
                        mk.set.dataset().add(v)
                        extrakeysdict[k] = mk
            except _NeedMoreKeysException:
                # Prepare the keys
                extra_keys[:] = list(set(itertools.chain((k for k,v in remove_uniquekeys if v.getkey() in updated_keyset or v.getkey() in autoremove_keys),
                                                         (k for k,v in remove_multikeys if v.getkey() in updated_keyset or v.getkey() in autoremove_keys),
                                                         (k for k,_ in update_uniquekeys),
                                                         (k for k,_ in update_multikeys))))
                extra_key_set[:] = list(set(itertools.chain((UniqueKeyReference.get_keyset_from_key(k) for k,v in remove_uniquekeys if v.getkey() in updated_keyset or v.getkey() in autoremove_keys),
                                                         (MultiKeyReference.get_keyset_from_key(k) for k,v in remove_multikeys if v.getkey() in updated_keyset or v.getkey() in autoremove_keys),
                                                         (UniqueKeyReference.get_keyset_from_key(k) for k,_ in update_uniquekeys),
                                                         (MultiKeyReference.get_keyset_from_key(k) for k,_ in update_multikeys))))
                auto_remove_keys.clear()
                auto_remove_keys.update(autoremove_keys.difference(keys[:orig_len])
                                                          .difference(extra_keys)
                                                          .difference(extra_key_set))
                raise
            else:
                extrakeys_list = list(extrakeysdict.items())
                extrakeyset_list = list(extrakeysetdict.items())
                autoremove_list = list(autoremove_keys.difference(updated_keys)
                                                      .difference(extrakeysdict.keys())
                                                      .difference(extrakeysetdict.keys()))
                return (tuple(itertools.chain(updated_keys,
                                              (k for k,_ in extrakeys_list),
                                              (k for k,_ in extrakeyset_list),
                                              autoremove_list)),
                        tuple(itertools.chain(updated_values,
                                               (v for _,v in extrakeys_list),
                                               (v for _,v in extrakeyset_list),
                                               [None] * len(autoremove_list))))
                        
        def object_updater(keys, values, timestamp):
            old_version = {}
            for k, v in zip(keys, values):
                if v is not None and hasattr(v, 'setkey'):
                    v.setkey(k)
                if v is not None and hasattr(v, 'kvdb_createtime'):
                    old_version[k] = (getattr(v, 'kvdb_createtime'), getattr(v, 'kvdb_updateversion', 1))
            updated_keys, updated_values = updater_with_key(keys, values, timestamp)
            updated_ref[0] = tuple(updated_keys)
            new_version = []
            for k,v in zip(updated_keys, updated_values):
                if v is None:
                    new_version.append((timestamp, -1))
                elif k in old_version:
                    ov = old_version[k]
                    setattr(v, 'kvdb_createtime', ov[0])
                    setattr(v, 'kvdb_updateversion', ov[1] + 1)
                    new_version.append((ov[0], ov[1] + 1))
                else:
                    setattr(v, 'kvdb_createtime', timestamp)
                    setattr(v, 'kvdb_updateversion', 1)
                    new_version.append((timestamp, 1))
            updated_ref[1] = new_version
            return (updated_keys, updated_values)
        start_time = self.apiroutine.scheduler.current_time
        retry_times = 1
        while True:
            try:
                await call_api(self.apiroutine, 'kvstorage', 'updateallwithtime',
                                 {'keys': keys + tuple(auto_remove_keys) + \
                                         tuple(extra_keys) + tuple(extra_key_set),
                                         'updater': object_updater})
            except _NeedMoreKeysException:
                if maxtime is not None and\
                    self.apiroutine.scheduler.current_time - start_time > maxtime:
                    raise TransactionTimeoutException
                retry_times += 1
            except Exception:
                self._logger.debug("Transaction %r interrupted in %r retries", updater, retry_times)
                raise
            else:
                self._logger.debug("Transaction %r done in %r retries", updater, retry_times)                
                break
        # Short cut update notification
        update_keys = self._watchedkeys.intersection(updated_ref[0])
        self._updatekeys.update(update_keys)
        for k,v in zip(updated_ref[0], updated_ref[1]):
            k = _str(k)
            if k in update_keys:
                v = tuple(v)
                oldv = self._update_version.get(k, (0, -1))
                if oldv < v:
                    self._update_version[k] = v
        if not self._requests:
            # Fake notification
            await self.apiroutine.wait_for_send(RetrieveRequestSend())
        await self._notifier.publish(updated_ref[0], updated_ref[1])
    
    async def gettimestamp(self):
        """
        Get a timestamp from database server
        """
        _timestamp = None
        def _updater(keys, values, timestamp):
            nonlocal _timestamp
            _timestamp = timestamp
            return ((), ())
        await call_api(self.apiroutine, 'kvstorage', 'updateallwithtime',
                                 {'keys': (),
                                  'updater': _updater})
        return _timestamp

    def watchlist(self, requestid = None):
        """
        Return a dictionary whose keys are database keys, and values are lists of request ids.
        Optionally filtered by request id
        """
        return dict((k,list(v)) for k,v in self._watches.items() if requestid is None or requestid in v)

    async def walk(self, keys, walkerdict, requestid, nostale = False):
        """
        Recursively retrieve keys with customized functions.
        walkerdict is a dictionary ``key->walker(key, obj, walk, save)``.
        """
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append((keys, rid, 'walk', dict(walkerdict), requestid))
        if notify:
            await self.apiroutine.wait_for_send(RetrieveRequestSend())
        ev = await RetrieveReply.createMatcher(rid)
        if hasattr(ev, 'exception'):
            raise ev.exception
        if nostale and ev.stale:
            raise StaleResultException(ev.result)
        return ev.result
    
    async def asynctransact(self, asyncupdater, withtime = False,
                            maxretry = None, maxtime=60):
        """
        Read-Write transaction with asynchronous operations.
        
        First, the `asyncupdater` is called with `asyncupdater(last_info, container)`.
        `last_info` is the info from last `AsyncTransactionLockException`.
        When `asyncupdater` is called for the first time, last_info = None.
        
        The async updater should be an async function, and return
        `(updater, keys)`. The `updater` should
        be a valid updater function used in `transaction` API. `keys` will
        be the keys used in the transaction.
        
        The async updater can return None to terminate the transaction
        without exception.
        
        After the call, a transaction is automatically started with the
        return values of `asyncupdater`.
        
        `updater` can raise `AsyncTransactionLockException` to restart
        the transaction from `asyncupdater`.
        
        :param asyncupdater: An async updater `asyncupdater(last_info, container)`
                             which returns `(updater, keys)`
        
        :param withtime: Whether the returned updater need a timestamp
        
        :param maxretry: Limit the max retried times
        
        :param maxtime: Limit the execution time. The transaction is abandoned
                        if still not completed after `maxtime` seconds. 
        """
        start_time = self.apiroutine.scheduler.current_time
        def timeleft():
            if maxtime is None:
                return None
            else:
                time_left = maxtime + start_time - \
                                self.apiroutine.scheduler.current_time
                if time_left <= 0:
                    raise TransactionTimeoutException
                else:
                    return time_left
        retry_times = 0
        last_info = None
        while True:
            timeout, r = \
                    await self.apiroutine.execute_with_timeout(
                            timeleft(),
                            asyncupdater(last_info, self.apiroutine)
                        )
            if timeout:
                raise TransactionTimeoutException
            if r is None:
                return
            updater, keys = r
            try:
                await self.transact(keys, updater, withtime, timeleft())
            except AsyncTransactionLockException as e:
                retry_times += 1
                if maxretry is not None and retry_times > maxretry:
                    raise TransactionRetryExceededException
                # Check time left
                timeleft()
                last_info = e.info
            except Exception:
                self._logger.debug("Async transaction %r interrupted in %r retries", asyncupdater, retry_times + 1)
                raise
            else:
                self._logger.debug("Async transaction %r done in %r retries", asyncupdater, retry_times + 1)
                break
    
    async def writewalk(self, keys, walker, withtime = False, maxtime = 60):
        """
        A read-write transaction with walkers
        
        :param keys: initial keys used in walk. Provide keys already known to
                     be necessary to optimize the transaction.

        :param walker: A walker should be `walker(walk, write)`,
                           where `walk` is a function `walk(key)->value`
                           to get a value from the database, and
                           `write` is a function `write(key, value)`
                           to save value to the database.
                           
                           A value can be write to a database any times.
                           A `walk` called after `write` is guaranteed
                           to retrieve the previously written value.

        :param withtime: if withtime=True, an extra timestamp parameter is given to
                         walkers, so walker should be
                         `walker(walk, write, timestamp)`
        
        :param maxtime: max execution time of this transaction
        """
        @functools.wraps(walker)
        async def _asyncwalker(last_info, container):
            return (keys, walker)
        return await self.asyncwritewalk(_asyncwalker, withtime, maxtime)

    async def asyncwritewalk(self, asyncwalker, withtime = False, maxtime = 60):
        """
        A read-write transaction with walker factory
        
        :param asyncwalker: an async function called as `asyncwalker(last_info, container)`
                            and returns (keys, walker), which
                            are the same as parameters of `writewalk`
        
                            :param keys: initial keys used in walk
                            
                            :param walker: A walker should be `walker(walk, write)`,
                                               where `walk` is a function `walk(key)->value`
                                               to get a value from the database, and
                                               `write` is a function `write(key, value)`
                                               to save value to the database.
                                               
                                               A value can be write to a database any times.
                                               A `walk` called after `write` is guaranteed
                                               to retrieve the previously written value.
                                               
                                               raise AsyncTransactionLockException in walkers
                                               to restart the transaction
        
        :param withtime: if withtime=True, an extra timestamp parameter is given to
                         walkers, so walkers should be
                         `walker(key, value, walk, write, timestamp)`
        
        :param maxtime: max execution time of this transaction
        """
        @functools.wraps(asyncwalker)
        async def _asyncupdater(last_info, container):
            if last_info is not None:
                from_walker, real_info = last_info
                if not from_walker:
                    keys, orig_keys, walker = real_info
                else:
                    r = await asyncwalker(real_info, container)
                    if r is None:
                        return None
                    keys, walker = r
                    orig_keys = keys
            else:
                r = await asyncwalker(None, container)
                if r is None:
                    return None
                keys, walker = r
                orig_keys = keys
            @functools.wraps(walker)
            def _updater(keys, values, timestamp):
                _stored_objs = dict(zip(keys, values))
                if self.debuggingupdater:
                    _stored_old_values = {k: v.jsonencode()
                                          for k,v in zip(keys, values)
                                          if hasattr(v, 'jsonencode')}
                # Keys written by walkers
                _walker_write_dict = {}
                _lost_keys = set()
                _used_keys = set()
                def _walk(key):
                    if key not in _stored_objs:
                        _lost_keys.add(key)
                        raise WalkKeyNotRetrieved(key)
                    else:
                        if key not in _walker_write_dict:
                            _used_keys.add(key)
                        return _stored_objs[key]
                def _write(key, value):
                    _walker_write_dict[key] = value
                    _stored_objs[key] = value
                try:
                    if withtime:
                        walker(_walk, _write, timestamp)
                    else:
                        walker(_walk, _write)
                except AsyncTransactionLockException as e:
                    raise AsyncTransactionLockException((True, e.info))
                if _lost_keys:
                    _lost_keys.update(_used_keys)
                    _lost_keys.update(orig_keys)
                    raise AsyncTransactionLockException((False, (_lost_keys, orig_keys, walker)))
                if self.debuggingupdater:
                    # Check if there are changes not written
                    for k, v in _stored_old_values.items():
                        if k not in _walker_write_dict:
                            v2 = _stored_objs[k]
                            assert hasattr(v2, 'jsonencode') and v2.jsonencode() == v
                if _walker_write_dict:
                    return tuple(zip(*_walker_write_dict.items()))
                else:
                    return (), ()
            return (_updater, keys)
        return await self.asynctransact(_asyncupdater, True, maxtime=maxtime)



'''
Created on 2015/06/01

:author: hubo
'''
from __future__ import print_function, absolute_import, division 
import warnings


class IsMatchExceptionWarning(Warning):
    pass


class EventMatcher(object):
    '''
    A matcher to match an event
    '''
    def __init__(self, indices, judgeFunc = None):
        # cut indices into real size
        for i in range(len(indices) - 1, -1, -1):
            if indices[i] is not None:
                break
        self.indices = indices[:i+1]
        if judgeFunc is not None:
            def _warning_judge(e):
                try:
                    return judgeFunc(e)
                except Exception as exc:
                    # Do not crash
                    warnings.warn(IsMatchExceptionWarning('Exception raised when _ismatch is calculated: %r. event = %r, matcher = %r, _ismatch = %r'
                                                          % (exc, e, self, judgeFunc)))
                    return False
            self.judge = _warning_judge
    def judge(self, event):
        return True
    def isMatch(self, event, indexStart = 0):
        if len(self.indices) > len(event.indices):
            return False
        for i in range(indexStart, len(self.indices)):
            if self.indices[i] is not None and self.indices[i] != event.indices[i]:
                return False
        return self.judge(event)    
    def __repr__(self):
        cls = type(self)
        return '<EventMatcher:' + \
            repr(self.indices) + '>'
    
    def __await__(self):
        """
        event = yield from matcher
        
        or 
        
        event = await matcher
        """
        ev, _ = yield (self,)
        return ev


class M_(object):
    """
    Awaitable object for multiple matchers
    
    ```
    event, matcher = await M_(matcher1, matcher2)
    ```
    """
    __slots__ = ('_matchers',)
    def __init__(self, *matchers):
        self._matchers = matchers
    
    def __await__(self):
        """
        ```
        event, matcher = yield from M_(matcher1, matcher2)
        ```
        
        equivalent to
        
        ```
        event, matcher = yield (matcher1, matcher2)
        ```
        """
        return (yield self._matchers)


class DiffRef_(object):
    """
    Append some matchers to a diff without breaking the difference structure
    """
    __slots__ = ('origin', 'add', 'result', 'length')
    def __init__(self, origin, add):
        self.origin = origin
        self.add = add
        self.result = None
        self.length = None
    
    def __add__(self, matchers):
        return DiffRef_(self.origin, self.add + matchers)
    
    def __iter__(self):
        if self.result is None:
            self.result = tuple(self.origin) + self.add
        return iter(self.result)
    
    def __len__(self):
        return len(self.origin) + len(self.add)
    
    def two_way_difference(self, b):
        """
        Return (self - b, b - self)
        """
        return self.origin.two_way_difference(b, self.add)

    def __await__(self):
        return (yield self)
    

class Diff_(object):
    """
    Special "differenced" set. Items in 'base', 'add', 'remove' must not be same
    
    Used by `wait_for_all`
    """
    __slots__ = ('base', 'add', 'remove', 'result', 'length')
    def __init__(self, base = (), add = (), remove = ()):
        self.base = base
        self.add = add
        self.remove = remove
        self.result = None
        self.length = None

    def __add__(self, matchers):
        return DiffRef_(self, matchers)
    
    def __iter__(self):
        if self.result is None:
            add = set()
            remove = set()
            base = self
            while isinstance(base, Diff_):
                if len(base) == len(base.add):
                    base = base.add
                    break
                add.update(base.add)
                remove.update(base.remove)
                base = base.base
            add.update(base)
            add.difference_update(remove)
            self.result = tuple(add)
        return iter(self.result)

    def __len__(self):
        if self.length is None:
            l = 0
            base = self
            while isinstance(base, Diff_):
                l += len(base.add) - len(base.remove)
                base = base.base
            l += len(base)
            self.length = l
            return l
        else:
            return self.length

    def two_way_difference(self, b, extra_add = (), extra_remove = ()):
        """
        Return (self - b, b - self)
        """
        if self is b:
            return ((), ())
        if isinstance(b, DiffRef_):
            extra_remove = extra_remove + b.add
            b = b.origin
        if extra_add == extra_remove:
            extra_add = extra_remove = ()
        if isinstance(b, Diff_):
            if self.base is b.base:
                first = self.add + b.remove
                second = self.remove + b.add
            elif self.base is b:
                first = self.add
                second = self.remove
            elif b.base is self:
                first = b.remove
                second = b.add
            else:
                first = self
                second = b
        else:
            first = self
            second = b
        if not first and not extra_add:
            return ((), tuple(second) + tuple(extra_remove))
        elif not second and not extra_remove:
            return (tuple(first) + tuple(extra_add), ())
        else:
            first = set(first)
            first.update(extra_add)
            second = set(second)
            second.update(extra_remove)
            return tuple(first.difference(second)), tuple(second.difference(first))

    def __await__(self):
        return (yield self)


def with_indices(*args):
    '''
    Create indices for an event class. Every event class must be decorated with this decorator.
    '''
    def decorator(cls):
        for c in cls.__bases__:
            if hasattr(c, '_indicesNames'):
                cls._classnameIndex = c._classnameIndex + 1
                for i in range(0, cls._classnameIndex):
                    setattr(cls, '_classname' + str(i), getattr(c, '_classname' + str(i)))
                setattr(cls, '_classname' + str(cls._classnameIndex), cls._getTypename())
                cls._indicesNames = c._indicesNames + ('_classname' + str(cls._classnameIndex),) + args
                cls._generateTemplate()
                return cls
        cls._classnameIndex = -1
        cls._indicesNames = args
        cls._generateTemplate()
        return cls
    return decorator


withIndices = with_indices


class Event(object):
    '''
    A generated event with indices
    '''
    canignore = True
    _indicesNames = ()
    _classnameIndex = -1
    def __init__(self, *args, **kwargs):
        '''
        :param args: index values like 12,"read",... content are type-depended.
        
        :param kwargs:
        
            *indices*
                input indices by name
                
            canignore
                if the event is not processed, whether it is safe to ignore the event.
                
                If it is not, the processing queue might be blocked to wait for a proper event processor.
                Default to True.
                
            *others*
                the properties will be set on the created event
        '''
        if kwargs and not args:
            indicesNames = self.indicesNames()
            indices = tuple(kwargs[k] if k[:10] != '_classname' else getattr(self, k) for k in indicesNames)
        else:
            indices = tuple(self._generateIndices(args))
        self.indices = indices
        if kwargs:
            self.__dict__.update(kwargs)

    def __getattr__(self, key):
        indicesNames = self.indicesNames()
        def _cache(v):
            setattr(self, key, v)
            return v
        try:
            i = indicesNames.index(key)
        except ValueError:
            raise AttributeError(key)
        else:
            return _cache(self.indices[i])

    @classmethod
    def indicesNames(cls):
        '''
        :returns: names of indices
        '''
        return getattr(cls, '_indicesNames', ())
    @classmethod
    def _getTypename(cls):
        module = cls.__module__
        if module is None:
            return cls.__name__
        else:
            return module  + '.' + cls.__name__        
    @classmethod
    def getTypename(cls):
        '''
        :returns: return the proper name to match
        '''
        if cls is Event:
            return None
        else:
            for c in cls.__bases__:
                if issubclass(c, Event):
                    if c is Event:
                        return cls._getTypename()
                    else:
                        return c.getTypename()
    @classmethod
    def _generateTemplate(cls):
        names = cls.indicesNames()
        template = [None] * len(names)
        argpos = []
        leastsize = 0
        for i in range(0, len(names)):
            if names[i][:10] == '_classname':
                template[i] = getattr(cls, names[i])
                leastsize = i + 1
            else:
                argpos.append(i)
        cls._template = template
        cls._argpos = argpos
        cls._leastsize = leastsize
    @classmethod
    def _generateIndices(cls, args):
        indices = cls._template[:]
        ap = cls._argpos
        lp = 0
        if args:
            for i in range(0, len(args)):
                indices[ap[i]] = args[i]
            lp = ap[len(args) - 1] + 1
        return indices[:max(cls._leastsize, lp)]
    @classmethod
    def createMatcher(cls, *args, **kwargs):
        '''
        :param _ismatch: user-defined function ismatch(event) for matching test
        :param \*args: indices
        :param \*\*kwargs: index_name=index_value for matching criteria
        '''
        if kwargs and not args:
            return EventMatcher(tuple(getattr(cls, ind) if ind[:10] == '_classname' else kwargs.get(ind) for ind in cls.indicesNames()), kwargs.get('_ismatch'))
        else:
            return EventMatcher(tuple(cls._generateIndices(args)), kwargs.get('_ismatch'))
    def __repr__(self):
        cls = type(self)
        return '<' + cls.__module__ + '.' + cls.__name__ + '(' + self.getTypename() + '): {' + \
            ', '.join(repr(k) + ': ' + repr(v) for k,v in zip(self.indicesNames(), self.indices)) + '}>'
    def canignorenow(self):
        '''
        Extra criteria for an event with canignore = False.
        When this event returns True, the event is safely ignored.
        '''
        return False




import logging
import copy

from uuid import uuid1
from vlcp.server.module import Module,depend,api,call_api
from vlcp.event.runnable import RoutineContainer
from vlcp.utils.networkmodel import *
from vlcp.utils.dataobject import watch_context,set_new,dump,ReferenceObject,\
    request_context, create_new
from vlcp.utils.ethernet import ip4_addr
from vlcp.utils.netutils import format_network_cidr,check_ip_address, parse_ip4_network, ip_in_network,\
    parse_ip4_address

import vlcp.service.kvdb.objectdb as objectdb
from vlcp.config.config import defaultconfig
from pychecktype.checked import checked
from pychecktype import tuple_
from vlcp.utils.typelib import ip_address_type, cidr_nonstrict_type
from collections import OrderedDict
from contextlib import suppress
from vlcp.utils.exceptions import WalkKeyNotRetrieved


@depend(objectdb.ObjectDB)
@defaultconfig
class VRouterApi(Module):
    """
    Standard network model for L3 SDN
    """
    def __init__(self,server):
        super(VRouterApi,self).__init__(server)

        self.app_routine = RoutineContainer(self.scheduler)
        self._reqid = 0

        self.createAPI(api(self.createvirtualrouter,self.app_routine),
                       api(self.createvirtualrouters,self.app_routine),
                       api(self.updatevirtualrouter,self.app_routine),
                       api(self.updatevirtualrouters,self.app_routine),
                       api(self.deletevirtualrouter,self.app_routine),
                       api(self.deletevirtualrouters,self.app_routine),
                       api(self.listvirtualrouters,self.app_routine),
                       api(self.addrouterinterface,self.app_routine),
                       api(self.addrouterinterfaces,self.app_routine),
                       api(self.removerouterinterface,self.app_routine),
                       api(self.removerouterinterfaces,self.app_routine),
                       api(self.listrouterinterfaces,self.app_routine))

    async def _dumpkeys(self, keys, filter=None):
        self._reqid += 1
        reqid = ('virtualrouter', self._reqid)

        with request_context(reqid, self.app_routine):
            retobjs = await call_api(self.app_routine,'objectdb','mget',{'keys':keys,'requestid':reqid})
            if filter is None:
                return [dump(o) for o in retobjs]
            else:
                return [dump(o) for o in retobjs if o is not None and all(getattr(o, k, None) == v for k, v in filter.items())]

    async def _dumpone(self, key, filter):
        return await self._dumpkeys([key], filter)

    async def createvirtualrouter(self, id: (str, None) = None,
                                        **kwargs: {"?routes": [tuple_((cidr_nonstrict_type, ip_address_type))]}):
        """
        Create a virtual router
        
        :param id: Virtual router id. If omitted, an UUID is generated.
        
        :param \*\*kwargs: extra attributes for creation.
        
        :return: A dictionary of information of the virtual router.
        """
        if not id:
            id = str(uuid1())

        router = {"id":id}
        router.update(kwargs)

        return await self.createvirtualrouters([router])
    
    @checked
    async def createvirtualrouters(self, routers: [{"?id": str,
                                                     "?routes": [tuple_((cidr_nonstrict_type, ip_address_type))]}]):
        """
        Create multiple virtual routers in a transaction
        """
        idset = set()
        newrouters = []
        for router in routers:
            router = copy.deepcopy(router)
            if "id" not in router:
                router["id"] = str(uuid1())
            else:
                if router["id"] in idset:
                    raise ValueError("Repeated ID: " + router['id'])
                else:
                    idset.add(router['id'])

            newrouters.append(router)
        
        routerkeys = [VRouter.default_key(r['id']) for r in newrouters]
        routersetkey = [VRouterSet.default_key()]

        routerobjects = [self._createvirtualrouter(**r) for r in newrouters ]

        def createrouterdb(keys,values):
            routerset = values[0]

            for i in range(0,len(routerkeys)):
                values[i+1] = set_new(values[i+1],routerobjects[i])
                routerset.set.dataset().add(routerobjects[i].create_weakreference())

            return keys,values
        await call_api(self.app_routine,"objectdb","transact",
                         {"keys":routersetkey+routerkeys,"updater":createrouterdb})
        return await self._dumpkeys(routerkeys)

    def _createvirtualrouter(self,id,**kwargs):

        router = VRouter.create_instance(id)

        # [(prefix, nexthop),(prefix, nexthop)]
        if 'routes' in kwargs:
            router.routes.extend(kwargs['routes'])

        for k,v in kwargs.items():
            if k != "routes":
                setattr(router,k,v)

        return router

    async def updatevirtualrouter(self, id: str, **kwargs: {"?routes": [tuple_((cidr_nonstrict_type, ip_address_type))]}):
        """
        Update virtual router
        """
        if not id:
            raise ValueError("must specify id")
        router = {"id":id}
        router.update(kwargs)
        return await self.updatevirtualrouters([router])
    
    @checked
    async def updatevirtualrouters(self, routers: [{"id": str,
                                                    "?routes": [tuple_((cidr_nonstrict_type, ip_address_type))]}]):
        "Update multiple virtual routers"
        idset = set()
        for router in routers:
            if 'id' not in router:
                raise ValueError("must specify id")
            else:
                if router['id'] in idset:
                    raise ValueError("Repeated ID: " + router['id'])
                else:
                    idset.add(router['id'])
        routerkeys = [VRouter.default_key(r['id']) for r in routers]

        def updaterouter(keys,values):

            for i in range(0,len(routers)):
                if values[i]:
                    for k,v in routers[i].items():
                        if k == 'routes':
                            values[i].routes[:] = copy.deepcopy(v)
                        else:
                            setattr(values[i], k, v)
                else:
                    raise ValueError("Virtual router not exists: " + routers[i]['id'])
            
            return keys,values
        await call_api(self.app_routine,'objectdb','transact',
                         {'keys':routerkeys,'updater':updaterouter})
        return await self._dumpkeys(routerkeys)

    async def deletevirtualrouter(self, id: str):
        "Delete virtual router"
        if not id:
            raise ValueError("must specify id")

        router = {"id":id}

        return await self.deletevirtualrouters([router])
    
    @checked
    async def deletevirtualrouters(self, routers: [{"id": str}]):
        "Delete multiple virtual routers"
        idset = set()
        for router in routers:
            if 'id' not in router:
                raise ValueError("must specify id")
            else:
                if router['id'] in idset:
                    raise ValueError("Repeated ID: " + router['id'])
                else:
                    idset.add(router['id'])

        routersetkey = [VRouterSet.default_key()]
        routerkeys = [VRouter.default_key(v['id']) for v in routers]

        def deletevrouter(keys,values):
            routerset = values[0]

            for i in range(0,len(routers)):
                if values[i+1].interfaces.dataset():
                    raise ValueError("there are still interface(s) in this router, delete them first")
                routerset.set.dataset().discard(values[i+1].create_weakreference())

            return keys,[routerset] + [None] * len(routers)
        await call_api(self.app_routine,"objectdb","transact",
                             {"keys":routersetkey + routerkeys,"updater":deletevrouter})
        return {"status":"OK"}

    async def listvirtualrouters(self, id: (str, None) = None, **kwargs):
        """
        Query virtual router
        
        :param id: if specified, only return virtual router with this ID
        
        :param \*\*kwargs: customized filter
        
        :return: a list of dictionaries each stands for a matched virtual router.
        """
        if id:
            return await self._dumpone(VRouter.default_key(id), kwargs)
        else:
            # we get all router from vrouterset index
            routersetkey = VRouterSet.default_key()

            self._reqid += 1
            reqid = ("virtualrouter",self._reqid)

            def set_walker(key,set,walk,save):

                for weakobj in set.dataset():
                    routerkey = weakobj.getkey()

                    try:
                        router = walk(routerkey)
                    except KeyError:
                        pass
                    else:
                        if all(getattr(router,k,None) == v for k,v in kwargs.items()):
                            save(routerkey)

            def walker_func(set_func):
                def walker(key,obj,walk,save):
                    
                    if obj is None:
                        return
                    
                    set_walker(key,set_func(obj),walk,save)

                return walker
            with request_context(reqid, self.app_routine):
                _, values = await call_api(self.app_routine,"objectdb","walk",
                                                 {"keys":[routersetkey],"walkerdict":{routersetkey:walker_func(lambda x:x.set)},
                                                  "requestid":reqid})
                return [dump(r) for r in values]

    async def addrouterinterface(self, router: str, subnet: str, id: (str, None) = None,
                                 **kwargs: {'?ip_address': ip_address_type}):
        """
        Connect virtual router to a subnet
        
        :param router: virtual router ID
        
        :param subnet: subnet ID
        
        :param id: router port ID
        
        :param \*\*kwargs: customized options
        
        :return: A dictionary of information of the created router port
        """
        if not id:
            id = str(uuid1())

        if not router:
            raise ValueError("must specify virtual router ID")

        if not subnet:
            raise ValueError("must specify subnet ID")

        interface = {"id":id,"router":router,"subnet":subnet}

        interface.update(kwargs)
        return await self.addrouterinterfaces([interface])
    
    @checked
    async def addrouterinterfaces(self, interfaces: [{"router": str,
                                                      "subnet": str,
                                                      "?id": str,
                                                      '?ip_address': ip_address_type}]):
        """
        Create multiple router interfaces
        """
        keys = set()
        parameter_dict = OrderedDict()
        for interface in interfaces:
            interface = copy.deepcopy(interface)
            if 'id' not in interface:
                interface['id'] = str(uuid1())
            key = RouterPort.default_key(interface['id'])
            if key in parameter_dict:
                raise ValueError("Repeated ID: "+interface['id'])
            parameter_dict[key] = interface
            keys.add(key)
            keys.add(VRouter.default_key(interface['router']))
            keys.add(SubNet.default_key(interface['subnet']))
            keys.add(SubNetMap.default_key(interface['subnet']))
        
        def walker(walk, write):
            for key, parameters in parameter_dict.items():
                with suppress(WalkKeyNotRetrieved):
                    value = walk(key)
                    value = create_new(RouterPort, value, parameters['id'])
                    subnet = walk(SubNet.default_key(parameters['subnet']))
                    if subnet is None:
                        raise ValueError("Subnet " + parameters['subnet'] + " not exists")
                    subnet_map = walk(SubNetMap.default_key(parameters['subnet']))
                    router = walk(VRouter.default_key(parameters['router']))
                    if router is None:
                        raise ValueError("Virtual router " + parameters['router'] + " not exists")
                    if hasattr(subnet, 'router'):
                        # normal subnet can only have one router
                        _, (rid,) = VRouter._getIndices(subnet.router.getkey())
                        raise ValueError("Subnet %r is already in virtual router %r", parameters['subnet'], rid)
                    if hasattr(subnet_map, 'routers'):
                        if router.create_weakreference() in subnet_map.routers.dataset():
                            raise ValueError("Subnet %r is already in virtual router %r", parameters['subnet'],
                                                                                          parameters['router'])
                    if 'ip_address' in parameters:
                        if getattr(subnet, 'isexternal', False):
                            raise ValueError("Cannot specify IP address when add external subnet to virtual router")
                        # Check IP address in CIDR
                        nip = parse_ip4_address(parameters['ip_address'])
                        cidr, prefix = parse_ip4_network(subnet.cidr)
                        if not ip_in_network(nip, cidr, prefix):
                            raise ValueError("IP address " + parameters['ip_address'] + " not in subnet CIDR")
                        # Check IP not allocated
                        if str(nip) in subnet_map.allocated_ips:
                            raise ValueError("IP address " + parameters['ip_address'] + " has been used")
                        else:
                            # Save to allocated map
                            subnet_map.allocated_ips[str(nip)] = (value.create_weakreference(), router.create_weakreference())
                            write(subnet_map.getkey(), subnet_map)
                    else:
                        # Use gateway
                        if not hasattr(subnet, 'gateway'):
                            raise ValueError("Subnet " + subnet.id + " does not have a gateway, IP address on router port must be specified explicitly")
                    if not hasattr(subnet_map, 'routers'):
                        subnet_map.routers = DataObjectSet()
                    subnet_map.routers.dataset().add(router.create_weakreference())
                    if not hasattr(subnet_map, 'routerports'):
                        subnet_map.routerports = {}
                    subnet_map.routerports[router.id] = value.create_weakreference()
                    write(subnet_map.getkey(), subnet_map)
                    if not getattr(subnet, 'isexternal', False):
                        subnet.router = value.create_weakreference()
                        write(subnet.getkey(), subnet)
                    # Save port to router
                    router.interfaces.dataset().add(value.create_weakreference())
                    write(router.getkey(), router)
                    value.router = router.create_reference()
                    value.subnet = subnet.create_reference()
                    for k, v in parameters.items():
                        if k not in ('router', 'subnet', 'id'):
                            setattr(value, k, v)
                    write(key, value)
        await call_api(self.app_routine,'objectdb','writewalk',
                       {"keys":keys, 'walker':walker})
        return await self._dumpkeys(parameter_dict)

    async def removerouterinterface(self, router: str, subnet: str):
        """
        Remote a subnet from the router
        
        :param router: virtual router ID
        
        :param subnet: subnet ID
        
        :return: ``{"status": "OK"}``
        """
        if not router:
            raise ValueError("must specify router id")

        if not subnet:
            raise ValueError("must specify subnet id")

        interface = {"router": router, "subnet": subnet}

        await self.removerouterinterfaces([interface])
    
    @checked
    async def removerouterinterfaces(self, interfaces: [{"router": str,
                                                   "subnet": str}]):
        """
        Remote multiple subnets from routers
        """
        for interface in interfaces:
            if 'router' not in interface:
                raise ValueError("must specify router ID")
            if "subnet" not in interface:
                raise ValueError("must specify subnet ID")
        
        keys = set()
        keys.update(VRouter.default_key(interface['router']) for interface in interfaces)
        keys.update(SubNet.default_key(interface['subnet']) for interface in interfaces)
        keys.update(SubNetMap.default_key(interface['subnet']) for interface in interfaces)
        
        def walker(walk, write):
            for interface in interfaces:
                with suppress(WalkKeyNotRetrieved):
                    router = walk(VRouter.default_key(interface['router']))
                    if router is None:
                        raise ValueError("Virtual router " + interface['router'] + " not exists")
                    subnet = walk(SubNet.default_key(interface['subnet']))
                    if subnet is None:
                        raise ValueError("Subnet " + interface['subnet'] + " not exists")
                    subnet_map = walk(SubNetMap.default_key(interface['subnet']))
                    if router.create_weakreference() not in subnet_map.routers.dataset():
                        raise ValueError("Subnet %r not in virtual router %r" % (subnet.id, router.id))
                    port = subnet_map.routerports.pop(router.id)
                    subnet_map.routers.dataset().discard(router.create_weakreference())
                    write(subnet_map.getkey(), subnet_map)
                    router.interfaces.dataset().discard(port)
                    write(router.getkey(), router)
                    if hasattr(subnet, 'router') and subnet.router == port:
                        delattr(subnet, 'router')
                        write(subnet.getkey(), subnet)
                    write(port.getkey(), None)

        await call_api(self.app_routine,"objectdb","writewalk",{"keys": keys,
                                                               "walker": walker})
        return {'status': 'OK'}

    async def listrouterinterfaces(self, id: str, **kwargs):
        """
        Query router ports from a virtual router
        
        :param id: virtual router ID
        
        :param \*\*kwargs: customized filters on router interfaces
        
        :return: a list of dictionaries each stands for a matched router interface
        """
        if not id:
            raise ValueError("must specify router id")

        routerkey = VRouter.default_key(id)

        self._reqid += 1
        reqid = ("virtualrouter", self._reqid)
        
        def set_walker(key, interfaces, walk, save):
            for weakobj in interfaces.dataset():
                vpkey = weakobj.getkey()
                try:
                    virtualport = walk(vpkey)
                except KeyError:
                    pass
                else:
                    if all(getattr(virtualport,k,None) == v for k,v in kwargs.items()):
                        save(vpkey)
        
        def walk_func(filter_func):
            def walker(key,obj,walk,save):
                if obj is None:
                    return
                set_walker(key,filter_func(obj),walk,save)
            return walker
        with request_context(reqid, self.app_routine):
            _, values = await call_api(self.app_routine,"objectdb","walk",
                                            {"keys":[routerkey],"walkerdict":{routerkey:walk_func(lambda x:x.interfaces)},
                                                "requestid":reqid})
            return [dump(r) for r in values]

    async def load(self,container):

        initkeys = [VRouterSet.default_key(),DVRouterForwardSet.default_key(),
                    DVRouterExternalAddressInfo.default_key()]

        def init(keys,values):
            
            if values[0] is None:
                values[0] = VRouterSet()

            if values[1] is None:
                values[1] = DVRouterForwardSet()

            if values[2] is None:
                values[2] = DVRouterExternalAddressInfo()
            
            return keys,values

        
        await call_api(container,"objectdb","transact",
                         {"keys":initkeys,"updater":init})
        
        await Module.load(self, container)
